/* ---------------------------------------------------------------------------
    Copyright (c) 2003-2013 Micrel, Inc.
    Copyright (c) 2015 Microchip Technology Inc.
   ---------------------------------------------------------------------------

    NdisDevice.c - NDIS driver device functions.

    Author      Date        Description
    THa         01/26/11    Ported from NDIS 5.0 driver with code from
                            Microsoft sample driver.
    THa         05/02/11    Added burst lengths.
    THa         05/11/11    Added UDP test limit.
    THa         11/04/13    Added OID_DEVICE_MODE to SupportedOids so that
                            Micrel applications work in Windows 7.
   ---------------------------------------------------------------------------
*/


#include "target.h"
#include "NdisDevice.h"


#pragma NDIS_PAGEABLE_FUNCTION(MiniportPause)
#pragma NDIS_PAGEABLE_FUNCTION(MiniportRestart)


#define OID_DEVICE_MODE  0xff0100cc

NDIS_OID SupportedOids[] =
{
    OID_GEN_HARDWARE_STATUS,
    OID_GEN_TRANSMIT_BUFFER_SPACE,
    OID_GEN_RECEIVE_BUFFER_SPACE,
    OID_GEN_TRANSMIT_BLOCK_SIZE,
    OID_GEN_RECEIVE_BLOCK_SIZE,
    OID_GEN_VENDOR_ID,
    OID_GEN_VENDOR_DESCRIPTION,
    OID_GEN_VENDOR_DRIVER_VERSION,
    OID_GEN_CURRENT_PACKET_FILTER,
    OID_GEN_CURRENT_LOOKAHEAD,
    OID_GEN_DRIVER_VERSION,
    OID_GEN_MAXIMUM_TOTAL_SIZE,

    OID_GEN_XMIT_OK,
    OID_GEN_RCV_OK,
    OID_GEN_XMIT_ERROR,
    OID_GEN_RCV_ERROR,
    OID_GEN_RCV_NO_BUFFER,
    OID_GEN_STATISTICS,

    OID_GEN_RCV_CRC_ERROR,
    OID_GEN_TRANSMIT_QUEUE_LENGTH,       // Optional

    OID_GEN_BYTES_RCV,
    OID_GEN_BYTES_XMIT,
    OID_GEN_RCV_DISCARDS,
    OID_GEN_XMIT_DISCARDS,

    OID_GEN_LINK_PARAMETERS,
    OID_GEN_INTERRUPT_MODERATION,

    OID_GEN_MEDIA_SUPPORTED,
    OID_GEN_MEDIA_IN_USE,
    OID_GEN_MAXIMUM_SEND_PACKETS,

    OID_802_3_PERMANENT_ADDRESS,
    OID_802_3_CURRENT_ADDRESS,
    OID_802_3_MULTICAST_LIST,
    OID_802_3_MAXIMUM_LIST_SIZE,

    OID_802_3_RCV_ERROR_ALIGNMENT,
    OID_802_3_XMIT_ONE_COLLISION,
    OID_802_3_XMIT_MORE_COLLISIONS,

    OID_802_3_XMIT_DEFERRED,             // Optional
    OID_802_3_XMIT_MAX_COLLISIONS,       // Optional
    OID_802_3_RCV_OVERRUN,               // Optional
    OID_802_3_XMIT_UNDERRUN,             // Optional
    OID_802_3_XMIT_LATE_COLLISIONS,      // Optional

    OID_PNP_CAPABILITIES,                // Optional
    OID_PNP_SET_POWER,
    OID_PNP_QUERY_POWER,
    OID_PNP_ADD_WAKE_UP_PATTERN,
    OID_PNP_REMOVE_WAKE_UP_PATTERN,
    OID_PNP_WAKE_UP_PATTERN_LIST,
    OID_PNP_ENABLE_WAKE_UP,

    OID_PNP_WAKE_UP_OK,
    OID_PNP_WAKE_UP_ERROR,

#if defined(NDIS620_MINIPORT)
    OID_PM_PARAMETERS,
    OID_PM_ADD_WOL_PATTERN,
    OID_PM_REMOVE_WOL_PATTERN,
    OID_PM_WOL_PATTERN_LIST,

    OID_RECEIVE_FILTER_ALLOCATE_QUEUE,
    OID_RECEIVE_FILTER_QUEUE_ALLOCATION_COMPLETE,
    OID_RECEIVE_FILTER_FREE_QUEUE,
    OID_RECEIVE_FILTER_CLEAR_FILTER,
    OID_RECEIVE_FILTER_SET_FILTER,
#endif

    OID_DEVICE_MODE,
};

/* -------------------------------------------------------------------------- */

/*
    GetConfigEntryUlong

Description:
    This function retrieves a double-word value from the registry.

Parameters:
    NDIS_HANDLE hRegistry
        Handle to registry key.

    NDIS_STRING* EntryString
        Pointer to NDIS string indicating the registry value.

    PUINT pulValue
        Buffer to store the value.

Return (BOOLEAN):
    TRUE if successful; otherwise FALSE.
*/

static
BOOLEAN GetConfigEntryUlong (
    NDIS_HANDLE  hRegistry,
    NDIS_STRING* EntryString,
    PUINT        pulValue )
{
    PNDIS_CONFIGURATION_PARAMETER pReturnedValue;
    NDIS_STATUS Status;
    NDIS_PARAMETER_TYPE parameter = NdisParameterHexInteger;

    NdisReadConfiguration(&Status, &pReturnedValue, hRegistry, EntryString,
        parameter);
    if (NDIS_STATUS_SUCCESS == Status)
    {
        *pulValue = pReturnedValue->ParameterData.IntegerData;
        return TRUE;
    }
    return FALSE;
}  // GetConfigEntryUlong


/*
    GetConfigEntryUshort

Description:
    This function retrieves a word value from the registry.

Parameters:
    NDIS_HANDLE hRegistry
        Handle to registry key.

    NDIS_STRING* EntryString
        Pointer to NDIS string indicating the registry value.

    PUSHORT pusValue
        Buffer to store the value.

Return (BOOLEAN):
    TRUE if successful; otherwise FALSE.
*/

static
BOOLEAN GetConfigEntryUshort (
    NDIS_HANDLE  hRegistry,
    NDIS_STRING* EntryString,
    PUSHORT      pusValue )
{
    PNDIS_CONFIGURATION_PARAMETER pReturnedValue;
    NDIS_STATUS Status;
    NDIS_PARAMETER_TYPE parameter = NdisParameterHexInteger;

    NdisReadConfiguration(&Status, &pReturnedValue, hRegistry, EntryString,
        parameter);
    if (NDIS_STATUS_SUCCESS == Status)
    {
        *pusValue = (USHORT) pReturnedValue->ParameterData.IntegerData;
        return TRUE;
    }
    return FALSE;
}  // GetConfigEntryUshort


/*
    GetConfigEntryUchar

Description:
    This function retrieves a byte value from the registry.

Parameters:
    NDIS_HANDLE hRegistry
        Handle to registry key.

    NDIS_STRING* EntryString
        Pointer to NDIS string indicating the registry value.

    PUCHAR pbValue
        Buffer to store the value.

Return (BOOLEAN):
    TRUE if successful; otherwise FALSE.
*/

static
BOOLEAN GetConfigEntryUchar (
    NDIS_HANDLE  hRegistry,
    NDIS_STRING* EntryString,
    PUCHAR       pbValue )
{
    PNDIS_CONFIGURATION_PARAMETER pReturnedValue;
    NDIS_STATUS Status;
    NDIS_PARAMETER_TYPE parameter = NdisParameterInteger;

    NdisReadConfiguration(&Status, &pReturnedValue, hRegistry, EntryString,
        parameter);
    if (NDIS_STATUS_SUCCESS == Status)
    {
        *pbValue = (UCHAR) pReturnedValue->ParameterData.IntegerData;
        return TRUE;
    }
    return FALSE;
}  // GetConfigEntryUchar


/*
    GetConfigEntryNum

Description:
    This function retrieves a double-word value from the registry.

Parameters:
    NDIS_HANDLE hRegistry
        Handle to registry key.

    NDIS_STRING* EntryString
        Pointer to NDIS string indicating the registry value.

    PUINT pulValue
        Buffer to store the value.

Return (BOOLEAN):
    TRUE if successful; otherwise FALSE.
*/

static
BOOLEAN GetConfigEntryNum (
    NDIS_HANDLE  hRegistry,
    NDIS_STRING* EntryString,
    PUINT        pulValue )
{
    PNDIS_CONFIGURATION_PARAMETER pReturnedValue;
    NDIS_STATUS Status;
    NDIS_PARAMETER_TYPE parameter = NdisParameterInteger;

    NdisReadConfiguration(&Status, &pReturnedValue, hRegistry, EntryString,
        parameter);
    if (NDIS_STATUS_SUCCESS == Status)
    {
        *pulValue = pReturnedValue->ParameterData.IntegerData;
        return TRUE;
    }
    return FALSE;
}  // GetConfigEntryNum


/*
    GetConfigEntryString

Description:
    This function retrieves a string value from the registry.

Parameters:
    NDIS_HANDLE hRegistry
        Handle to registry key.

    NDIS_STRING* EntryString
        Pointer to NDIS string indicating the registry value.

    PCHAR szValue
        Buffer to store the string value.

    USHORT wLength
        The length of the buffer.

Return (int):
    The length of the string.
*/

static
int GetConfigEntryString (
    NDIS_HANDLE  hRegistry,
    NDIS_STRING* EntryString,
    PCHAR        szValue,
    USHORT       wLength )
{
    PNDIS_CONFIGURATION_PARAMETER pReturnedValue;
    NDIS_STATUS Status;
    ANSI_STRING anszString;

    anszString.MaximumLength = wLength;
    anszString.Buffer = szValue;

    NdisReadConfiguration(&Status, &pReturnedValue, hRegistry, EntryString,
        NdisParameterString);
    if (NDIS_STATUS_SUCCESS == Status)
    {
        NdisUnicodeStringToAnsiString(&anszString,
            &pReturnedValue->ParameterData.StringData);
        return(anszString.Length);
    }
    return 0;
}  // GetConfigEntryString

/* -------------------------------------------------------------------------- */

#define nszcBusType                 NDIS_STRING_CONST("BusType")
#define nszcIoBaseAddress           NDIS_STRING_CONST("IoBaseAddress")
#define nszcInterruptNumber         NDIS_STRING_CONST("InterruptNumber")
#define nszcIoBaseLength            NDIS_STRING_CONST("IoLen")
#define nszcMemoryAddress           NDIS_STRING_CONST("MemoryBaseAddress")
#define nszcMemorySize              NDIS_STRING_CONST("MemorySize")
#define nszcSlotNumber              NDIS_STRING_CONST("SlotNumber")

#define nszcBurstRx                 NDIS_STRING_CONST("RxBurst")
#define nszcBurstTx                 NDIS_STRING_CONST("TxBurst")
#define nszcChecksumRx              NDIS_STRING_CONST("ChecksumRx")
#define nszcChecksumTx              NDIS_STRING_CONST("ChecksumTx")
#define nszcNetworkAddress          NDIS_STRING_CONST("NetworkAddress")
#define nszcSpeedDuplex             NDIS_STRING_CONST("SpeedDuplex")
#define nszcUdpTestLimit            NDIS_STRING_CONST("UdpTestLimit")

#define DEFAULT_IO_BASE_ADDRESS     0x300
#define DEFAULT_IO_BASE_LENGTH      0x10
#define DEFAULT_MEMORY_SIZE         0x1000
#define DEFAULT_MEMORY_WINDOW       0xd4000
#define DEFAULT_INTERRUPT_NUMBER    3
#define MAX_IOBASEADDR              0xFFFFFFFF
#define MIN_IOBASEADDR              0x0100
#define MAX_IRQ                     31
#define MIN_IRQ                     2


/*
    GetResources

Description:
    This function gets adapter resources from the operating system and the
    registry.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    NDIS_HANDLE hRegistry
        Handle to registry key.

    PNDIS_MINIPORT_INIT_PARAMETERS InitParameters
        Init parameters from driver.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

static
NDIS_STATUS GetResources (
        PNDIS_ADAPTER                  pAdapter,
    IN  NDIS_HANDLE                    hRegistry,
    IN  PNDIS_MINIPORT_INIT_PARAMETERS InitParameters )
{
    PNDIS_RESOURCE_LIST pResource = (PNDIS_RESOURCE_LIST)
        InitParameters->AllocatedResources;
    PCM_PARTIAL_RESOURCE_DESCRIPTOR pResDesc;
    UINT ulIndex;

    if (pResource)
    {
        // Parse the NDIS_RESOURCE_LIST (CM_PARTIAL_RESOURCE_LIST/
        // CM_PARTIAL_RESOURCE_DESCRIPTOR).
        for (ulIndex = 0; ulIndex < pResource->Count; ulIndex++)
        {
            pResDesc = &pResource->PartialDescriptors[ulIndex];
            // We handle 3 types of resources: I/O port, interrupt, and
            // memory base.  Ignore all others.
            switch (pResDesc->Type)
            {
                case CmResourceTypePort:
                        pAdapter->ulIOBaseAddress =
                            NdisGetPhysicalAddressLow(pResDesc->u.Port.Start);
                        pAdapter->ulIOLength = pResDesc->u.Port.Length;

#ifdef DBG
DbgPrint("  io: %x, %x"NEWLINE, pAdapter->ulIOBaseAddress,
    pAdapter->ulIOLength);
#endif
                    break;

                case CmResourceTypeInterrupt:
                    pAdapter->ulInterruptNumber = pResDesc->u.Interrupt.Vector;

#ifdef DBG
DbgPrint("  intr: %d"NEWLINE, pAdapter->ulInterruptNumber);
#endif
                    break;

                case CmResourceTypeMemory:
#ifdef _WIN64
                    pAdapter->MemoryAddress =
                        pResDesc->u.Memory.Start.QuadPart;
#else
                    pAdapter->MemoryAddress = (ULONG)
                        pResDesc->u.Memory.Start.u.LowPart;
#endif
                    pAdapter->ulMemorySize = pResDesc->u.Memory.Length;

#ifdef DBG
DbgPrint("  mem: %lx, %x"NEWLINE, pAdapter->MemoryAddress,
    pAdapter->ulMemorySize);
#endif
                    break;
            }
        }
    }
    else
    {
        NDIS_STRING nszIoBaseAddress = nszcIoBaseAddress;
        NDIS_STRING nszIoBaseLength = nszcIoBaseLength;
        NDIS_STRING nszInterruptNumber = nszcInterruptNumber;
        NDIS_STRING nszMemoryAddress = nszcMemoryAddress;
        NDIS_STRING nszMemorySize = nszcMemorySize;
        UINT        uiAddr;

        pAdapter->ulIOBaseAddress = DEFAULT_IO_BASE_ADDRESS;
        GetConfigEntryUlong(hRegistry, &nszIoBaseAddress,
            &pAdapter->ulIOBaseAddress);
        if (pAdapter->ulIOBaseAddress < MIN_IOBASEADDR ||
                pAdapter->ulIOBaseAddress > MAX_IOBASEADDR)
        {
            NdisWriteErrorLogEntry(pAdapter->hAdapter,
                NDIS_ERROR_CODE_UNSUPPORTED_CONFIGURATION, 1,
                pAdapter->ulIOBaseAddress);
            return NDIS_STATUS_FAILURE;
        }

        pAdapter->ulIOLength = DEFAULT_IO_BASE_LENGTH;
        GetConfigEntryUlong(hRegistry, &nszIoBaseLength,
            &pAdapter->ulIOLength);

        pAdapter->ulInterruptNumber = DEFAULT_INTERRUPT_NUMBER;
        GetConfigEntryUlong(hRegistry, &nszInterruptNumber,
            &pAdapter->ulInterruptNumber);
        if (pAdapter->ulInterruptNumber < MIN_IRQ ||
                pAdapter->ulInterruptNumber > MAX_IRQ)
        {
            NdisWriteErrorLogEntry(pAdapter->hAdapter,
                NDIS_ERROR_CODE_UNSUPPORTED_CONFIGURATION, 1,
                pAdapter->ulInterruptNumber);
            return NDIS_STATUS_FAILURE;
        }

        pAdapter->MemoryAddress = DEFAULT_MEMORY_WINDOW;
        GetConfigEntryUlong(hRegistry, &nszMemoryAddress, &uiAddr);
        pAdapter->MemoryAddress = uiAddr;

        pAdapter->ulMemorySize = DEFAULT_MEMORY_SIZE;
        GetConfigEntryUlong(hRegistry, &nszMemorySize,
            &pAdapter->ulMemorySize);
    }
    return NDIS_STATUS_SUCCESS;
}  // GetResources


/*
    ReadConfiguration

Description:
    This function gets configuration for the adapter from the operating
    system and the registry.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNDIS_MINIPORT_INIT_PARAMETERS InitParameters
        Init parameters from driver.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

static
NDIS_STATUS ReadConfiguration (
    IN  PNDIS_ADAPTER                  pAdapter,
    IN  PNDIS_MINIPORT_INIT_PARAMETERS InitParameters )
{
    // Handle for reading from registry.
    NDIS_HANDLE hRegistry;
    NDIS_STATUS Status;
    BOOLEAN bStatus;
    PHARDWARE pHardware = &pAdapter->Hardware;
    NDIS_STRING nszBusType = nszcBusType;
    NDIS_STRING nszAddress = nszcNetworkAddress;
    NDIS_STRING nszBurstRx = nszcBurstRx;
    NDIS_STRING nszBurstTx = nszcBurstTx;
    NDIS_STRING nszChecksumRx = nszcChecksumRx;
    NDIS_STRING nszChecksumTx = nszcChecksumTx;
    NDIS_STRING nszSpeed = nszcSpeedDuplex;
    NDIS_STRING nszUdpTestLimit = nszcUdpTestLimit;
    ULONG ulData;
    USHORT wData;
    UCHAR bData;
    PUCHAR NetworkAddress;
    UINT Length;

    // Open the configuration space.
    NDIS_CONFIGURATION_OBJECT Parameters;

    //
    // Open the registry for this adapter to read advanced
    // configuration parameters stored by the INF file.
    //

    NdisZeroMemory(&Parameters, sizeof(Parameters));

    Parameters.Header.Type = NDIS_OBJECT_TYPE_CONFIGURATION_OBJECT;
    Parameters.Header.Size = NDIS_SIZEOF_CONFIGURATION_OBJECT_REVISION_1;
    Parameters.Header.Revision = NDIS_CONFIGURATION_OBJECT_REVISION_1;

    Parameters.NdisHandle = pAdapter->hAdapter;
    Parameters.Flags = 0;

    Status = NdisOpenConfigurationEx(&Parameters, &hRegistry);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        NdisWriteErrorLogEntry(pAdapter->hAdapter,
            NDIS_ERROR_CODE_UNSUPPORTED_CONFIGURATION, 0);
        return Status;
    }

    // get bus type
    pAdapter->BusType = NdisInterfacePci;
    bStatus = GetConfigEntryUlong(hRegistry, &nszBusType,
        (PUINT) &pAdapter->BusType);

#ifdef DBG
    if (!bStatus)
        DbgPrint(" def");
DbgPrint("  bustype: %x"NEWLINE, pAdapter->BusType);
#endif

    bStatus = GetConfigEntryUchar(hRegistry, &nszBurstRx, &bData);
    if (bStatus)
        pHardware->m_dwRxBurstSize = bData;
    else
        pHardware->m_dwRxBurstSize = PBL_DEFAULT;
    bStatus = GetConfigEntryUchar(hRegistry, &nszBurstTx, &bData);
    if (bStatus)
        pHardware->m_dwTxBurstSize = bData;
    else
        pHardware->m_dwTxBurstSize = PBL_DEFAULT;
#if DBG
    DBG_PRINT("Burst = %u %u"NEWLINE,
        pHardware->m_dwRxBurstSize, pHardware->m_dwTxBurstSize);
#endif

    bStatus = GetConfigEntryUshort(hRegistry, &nszChecksumRx, &wData);
    if (bStatus)
    {
        if (wData)
        {
            pAdapter->ulRxChecksum = TASK_OFFLOAD_CHECKSUM;
        }
        else
            pAdapter->ulRxChecksum = 0;
    }
    bStatus = GetConfigEntryUshort(hRegistry, &nszChecksumTx, &wData);
    if (bStatus)
    {
        if (wData)
        {
            pAdapter->ulTxChecksum = TASK_OFFLOAD_CHECKSUM;
        }
        else
            pAdapter->ulTxChecksum = 0;
    }
#if DBG
    DBG_PRINT("Checksum = %u %u"NEWLINE,
        pAdapter->ulRxChecksum, pAdapter->ulTxChecksum);
#endif

    NdisReadNetworkAddress(&Status, &NetworkAddress, &Length, hRegistry);
    if ((Status == NDIS_STATUS_SUCCESS) && (Length == MAC_ADDRESS_LENGTH))
    {
#if 0
        ETH_NETWORK_COPY_ADDR(pHardware->m_bOverrideAddress, NetworkAddress);
#endif
        NdisMoveMemory(pHardware->m_bOverrideAddress, NetworkAddress,
            MAC_ADDRESS_LENGTH);
        pHardware->m_bMacOverrideAddr = TRUE;
    }

    bStatus = GetConfigEntryUlong(hRegistry, &nszSpeed, &ulData);
    if (bStatus)
    {
        if (ulData)
        {
            pHardware->m_bDuplex = (UCHAR)(ulData & 3);
            if (ulData >= 0x10000)
            {
                pHardware->m_bForceLink = 1;
                ulData -= 0x10000;
            }
            if (ulData >= 0x100)
            {
                pHardware->m_bSpeed = 100;
            }
            else if (ulData >= 0x10)
            {
                pHardware->m_bSpeed = 10;
            }
        }
#if DBG
        DBG_PRINT("SpeedDuplex = %x; %u %u %u"NEWLINE, ulData,
            pHardware->m_bSpeed, pHardware->m_bDuplex,
            pHardware->m_bForceLink);
#endif
    }

    bStatus = GetConfigEntryNum(hRegistry, &nszUdpTestLimit,
        &pAdapter->ulUdpTestLimit);
#if DBG
    DBG_PRINT("Udp Test Limit = %u"NEWLINE, pAdapter->ulUdpTestLimit);
#endif

    pHardware->m_RxDescInfo.cnAlloc = NUM_OF_RX_DESC;
    pHardware->m_TxDescInfo.cnAlloc = NUM_OF_TX_DESC;
    pHardware->m_pPciCfg = pAdapter;
    pHardware->m_pDevice = pAdapter;

    if ((Status = GetResources(pAdapter, hRegistry, InitParameters)) !=
            NDIS_STATUS_SUCCESS)
    {
        goto ReadConfigurationDone;
    }

ReadConfigurationDone:
    NdisCloseConfiguration(hRegistry);

    return Status;
}  // ReadConfiguration


/*
    DeRegisterAdapter

Description:
    This routine de-registers the adapter from the operating system.  It
    releases the I/O address range used by the adapter.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

VOID DeRegisterAdapter (
    PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;

#ifdef DBG
DbgPrint("DeRegister"NEWLINE);
#endif

    if (pHardware->m_pVirtualMemory)
    {
        NdisMUnmapIoSpace(pAdapter->hAdapter,
            pHardware->m_pVirtualMemory,
            pAdapter->ulMemorySize);
        pHardware->m_pVirtualMemory = NULL;
    }
    if (pHardware->m_ulVIoAddr)
    {
        NdisMDeregisterIoPortRange(pAdapter->hAdapter,
            pAdapter->ulIOBaseAddress, pAdapter->ulIOLength,
            (PVOID) pHardware->m_ulVIoAddr);
        pHardware->m_ulVIoAddr = 0;
    }
}  // DeRegisterAdapter


/*
    RegisterAdapter

Description:
    This function registers the adapter to the NDIS wrapper.  It acquires
    the I/O address range to run the adapter.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

NDIS_STATUS RegisterAdapter (
    PNDIS_ADAPTER pAdapter )
{
    NDIS_PHYSICAL_ADDRESS PhysicalAddress;
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    PHARDWARE pHardware = &pAdapter->Hardware;

#ifdef DBG
DbgPrint("Register"NEWLINE);
#endif

    if (pAdapter->ulMemorySize)
    {
#ifdef _WIN64
        PhysicalAddress.QuadPart = pAdapter->MemoryAddress;
#else
        NdisSetPhysicalAddressLow(PhysicalAddress,
            pAdapter->MemoryAddress);
        NdisSetPhysicalAddressHigh(PhysicalAddress, 0);
#endif
        Status = NdisMMapIoSpace((PVOID *) &pHardware->m_pVirtualMemory,
            pAdapter->hAdapter, PhysicalAddress,
            pAdapter->ulMemorySize);
        if (Status != NDIS_STATUS_SUCCESS)
        {
#if 0
            NdisWriteErrorLogEntry(pAdapter->hAdapter,
                NDIS_ERROR_CODE_RESOURCE_CONFLICT,
                1,
                ERRLOG_MAP_IO_SPACE);
#else
            NdisWriteErrorLogEntry(pAdapter->hAdapter,
                NDIS_ERROR_CODE_RESOURCE_CONFLICT, 0);
#endif
            goto RegisterAdapterError;
        }
    }
    if (pAdapter->ulIOLength)
    {
        Status = NdisMRegisterIoPortRange((PVOID *) &pHardware->m_ulVIoAddr,
            pAdapter->hAdapter, pAdapter->ulIOBaseAddress,
            pAdapter->ulIOLength);
        if (Status != NDIS_STATUS_SUCCESS)
        {
            NdisWriteErrorLogEntry(pAdapter->hAdapter,
                NDIS_ERROR_CODE_BAD_IO_BASE_ADDRESS, 0);
            goto RegisterAdapterError;
        }
    }

    return Status;

RegisterAdapterError:
    DeRegisterAdapter(pAdapter);
    return Status;
}  // RegisterAdapter

/* -------------------------------------------------------------------------- */

PVOID AllocMemory (
    IN  NDIS_HANDLE hAdapter,
    IN  ULONG       Size )
{
    return NdisAllocateMemoryWithTagPriority(hAdapter, Size, MEM_TAG,
        LowPoolPriority);
}  // AllocMemory


static
NDIS_STATUS AllocNetBufferListPool (
    IN  PNDIS_ADAPTER pAdapter )
{
    NET_BUFFER_LIST_POOL_PARAMETERS Parameters;

    NdisZeroMemory(&Parameters, sizeof(Parameters));

    Parameters.Header.Type = NDIS_OBJECT_TYPE_DEFAULT;
    Parameters.Header.Revision = NET_BUFFER_LIST_POOL_PARAMETERS_REVISION_1;
    Parameters.Header.Size =
        NDIS_SIZEOF_NET_BUFFER_LIST_POOL_PARAMETERS_REVISION_1;

    Parameters.ProtocolId = 0;
    Parameters.fAllocateNetBuffer = TRUE;
    Parameters.ContextSize = 0;
    Parameters.PoolTag = MEM_TAG;
    Parameters.DataSize = 0;

    pAdapter->hNetBufListPool = NdisAllocateNetBufferListPool(
        pAdapter->hAdapter, &Parameters);
    if (!pAdapter->hNetBufListPool)
        return NDIS_STATUS_RESOURCES;
    return NDIS_STATUS_SUCCESS;
}  // AllocNetBufferListPool


static
NDIS_STATUS AllocScatterGatherDma (
    IN  PNDIS_ADAPTER pAdapter )
{
    NDIS_STATUS Status;
    NDIS_SG_DMA_DESCRIPTION Description;
    PTCB_INFO pInfo = &pAdapter->TcbInfo;

    NdisZeroMemory(&Description, sizeof(Description));

    Description.Header.Type = NDIS_OBJECT_TYPE_SG_DMA_DESCRIPTION;
    Description.Header.Revision = NDIS_SG_DMA_DESCRIPTION_REVISION_1;
    Description.Header.Size = sizeof(NDIS_SG_DMA_DESCRIPTION);

    Description.Flags = 0;

    Description.MaximumPhysicalMapping = NDIS_BUFFER_SIZE;

    Description.ProcessSGListHandler = MiniportProcessSGList;

    Status = NdisMRegisterScatterGatherDma(pAdapter->hAdapter,
        &Description, &pInfo->hDma);
    if (Status == NDIS_STATUS_SUCCESS)
    {
        pInfo->ScatterGatherSize = Description.ScatterGatherListSize;
        pInfo->ScatterGatherBuffer = AllocMemory(pAdapter->hAdapter,
            pInfo->ScatterGatherSize * pInfo->cnAlloc);
        if (!pInfo->ScatterGatherBuffer)
        {
            return NDIS_STATUS_RESOURCES;
        }
    }
    return Status;
}  // AllocScatterGatherDma


/*
    InitTcb

Description:
    This local routine initializes the map register information structure.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

static void InitTcb (
        PNDIS_ADAPTER pAdapter )
{
    PTCB_INFO pInfo = &pAdapter->TcbInfo;
    PTCB pTcb;
    int i;

    pTcb = pInfo->TcbArray;
    for (i = 0; i < pInfo->cnAlloc; i++)
    {
        pTcb->ScatterGatherBuffer = pInfo->ScatterGatherBuffer + i *
            pInfo->ScatterGatherSize;
        pTcb->pAdapter = pAdapter;
        NdisInterlockedPushEntrySList(&pInfo->listHead,
            &pTcb->listEntry, &pInfo->listLock);
        pTcb++;
    }
}  // InitTcb


/*
    AllocateTcb

Description:
    This local function allocates the mapping registers for use in
    transmission.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

static NDIS_STATUS AllocateTcb (
        PNDIS_ADAPTER pAdapter )
{
    PTCB_INFO pInfo = &pAdapter->TcbInfo;
    PHARDWARE pHardware = &pAdapter->Hardware;
    NDIS_STATUS Status;

    pInfo->cnAlloc = pHardware->m_TxDescInfo.cnAlloc;
    Status = AllocScatterGatherDma(pAdapter);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }

    pInfo->TcbArray = AllocMemory(pAdapter->hAdapter,
        sizeof(TCB) * pInfo->cnAlloc);
    if (!pInfo->TcbArray)
    {
        return NDIS_STATUS_RESOURCES;
    }

    NdisAllocateSpinLock(&pInfo->listLock);
    NdisInitializeSListHead(&pInfo->listHead);
    InitTcb(pAdapter);

    return NDIS_STATUS_SUCCESS;
}  // AllocateTcb


/*
    AllocateSoftDescriptors

Description:
    This local function allocates software descriptors for manipulation in
    memory.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PTDescInfo pDescInfo
        Pointer to descriptor information structure.

    int fTransmit
        Indication that descriptors are for transmit.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

static NDIS_STATUS AllocateSoftDescriptors (
        NDIS_HANDLE hAdapter,
        PTDescInfo  pDescInfo,
        int         fTransmit )
{
    pDescInfo->pRing = AllocMemory(hAdapter,
        sizeof(TDesc) * pDescInfo->cnAlloc);
    if (!pDescInfo->pRing)
    {
        return NDIS_STATUS_FAILURE;
    }
    NdisZeroMemory(pDescInfo->pRing, sizeof(TDesc) * pDescInfo->cnAlloc);
    HardwareInitDescriptors(pDescInfo, fTransmit);

    return NDIS_STATUS_SUCCESS;
}  /* AllocateSoftDescriptors */


/*
    AllocateDescriptors

Description:
    This local function allocates hardware descriptors for receiving and
    transmitting.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

static NDIS_STATUS AllocateDescriptors (
        PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    int nOffset;
    NDIS_STATUS Status;

    /* Allocate memory for RX & TX descriptors. */
    pAdapter->DescPool.ulAllocSize =
        pHardware->m_RxDescInfo.nSize * pHardware->m_RxDescInfo.cnAlloc +
        pHardware->m_TxDescInfo.nSize * pHardware->m_TxDescInfo.cnAlloc +
        DESC_ALIGNMENT;
    pAdapter->DescPool.bCached = FALSE;

    NdisMAllocateSharedMemory(pAdapter->hAdapter,
        pAdapter->DescPool.ulAllocSize,
        pAdapter->DescPool.bCached,
        &pAdapter->DescPool.pAllocVirtual,
        &pAdapter->DescPool.AllocPhysicalAddr);

    if (pAdapter->DescPool.pAllocVirtual == NULL)
    {
        pAdapter->DescPool.ulAllocSize = 0;
        return NDIS_STATUS_RESOURCES;
    }
    else if (NdisGetPhysicalAddressHigh(
            pAdapter->DescPool.AllocPhysicalAddr) != 0)
    {
        NdisMFreeSharedMemory(pAdapter->hAdapter,
            pAdapter->DescPool.ulAllocSize,
            pAdapter->DescPool.bCached,
            pAdapter->DescPool.pAllocVirtual,
            pAdapter->DescPool.AllocPhysicalAddr);
        pAdapter->DescPool.ulAllocSize = 0;
        pAdapter->DescPool.pAllocVirtual = NULL;
        return NDIS_STATUS_RESOURCES;
    }

    NdisZeroMemory(pAdapter->DescPool.pAllocVirtual,
        pAdapter->DescPool.ulAllocSize);

    /* Align to the next cache line boundary. */
    nOffset = (((ULONG_PTR) pAdapter->DescPool.pAllocVirtual % DESC_ALIGNMENT)
        ? (DESC_ALIGNMENT - (int)
        ((ULONG_PTR) pAdapter->DescPool.pAllocVirtual % DESC_ALIGNMENT)) : 0);
    pAdapter->DescPool.pVirtual = pAdapter->DescPool.pAllocVirtual +
        nOffset;
    pAdapter->DescPool.ulPhysical = NdisGetPhysicalAddressLow(
        pAdapter->DescPool.AllocPhysicalAddr) + nOffset;

    /* Allocate receive/transmit descriptors. */
    pHardware->m_RxDescInfo.phwRing = (PTHw_Desc)
        pAdapter->DescPool.pVirtual;
    pHardware->m_RxDescInfo.ulRing = pAdapter->DescPool.ulPhysical;
    nOffset = pHardware->m_RxDescInfo.cnAlloc * pHardware->m_RxDescInfo.nSize;
    pHardware->m_TxDescInfo.phwRing = (PTHw_Desc)
        (pAdapter->DescPool.pVirtual + nOffset);
    pHardware->m_TxDescInfo.ulRing = pAdapter->DescPool.ulPhysical + nOffset;

    Status = AllocateSoftDescriptors(pAdapter->hAdapter,
        &pHardware->m_RxDescInfo, FALSE);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }
    Status = AllocateSoftDescriptors(pAdapter->hAdapter,
        &pHardware->m_TxDescInfo, TRUE);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }

    return NDIS_STATUS_SUCCESS;
}  /* AllocateDescriptors */


/*
    InitBuffers

Description:
    This local routine initializes the DMA buffer information structure.

Parameters:
    PBUFFER_INFO pBufInfo
        Pointer to DMA buffer information structure.

Return (None):
*/

static void InitBuffers (
    PBUFFER_INFO pBufInfo )
{
    PDMA_BUFFER pDma;
    int         i;

    pDma = pBufInfo->BufArray;
    for (i = 0; i < pBufInfo->cnAlloc; i++)
    {
        NdisInterlockedPushEntrySList(&pBufInfo->listHead, &pDma->listEntry,
            &pBufInfo->listLock);
        pDma++;
    }
}  /* InitBuffers */


/*
    AllocateBuffers

Description:
    This local function allocates DMA buffers for use in receiving and
    transmitting.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PBUFFER_INFO pBufInfo
        Pointer to DMA buffer information structure.

    int fPacket
        Indication that packets are allocated for receiving.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

static NDIS_STATUS AllocateBuffers (
    PNDIS_ADAPTER pAdapter,
    PBUFFER_INFO  pBufInfo,
    int           fPacket )
{
    NDIS_PHYSICAL_ADDRESS AllocPhysicalAddr;
    NDIS_PHYSICAL_ADDRESS PhysicalAddr;
    NDIS_STATUS Status;
    PDMA_BUFFER pDma;
    PUCHAR pAllocVirtual;
    PUCHAR pVirtual;
    UINT ulAllocSize;
    UINT ulPhysical;
    int i;
    int nOffset;
    PSHARED_MEM pBufPool = &pBufInfo->MemPool;
    NDIS_PHYSICAL_ADDRESS HighestAcceptableMax =
        NDIS_PHYSICAL_ADDRESS_CONST(-1, -1);

    /* Attempt to allocate all the buffer space in one big block. */
    /* If it fails, allocate each buffer individually. */
    pBufPool->ulAllocSize = (pBufInfo->cnAlloc * NDIS_BUFFER_SIZE) +
        BUFFER_ALIGNMENT;
    pBufPool->bCached = TRUE;

    NdisMAllocateSharedMemory(pAdapter->hAdapter,
        pBufPool->ulAllocSize,
        pBufPool->bCached,
        &pBufPool->pAllocVirtual,
        &pBufPool->AllocPhysicalAddr);

    /* Big buffer not allocated. */
    if (pBufPool->pAllocVirtual == NULL)
    {
        pBufPool->ulAllocSize = 0;
    }
    else if (NdisGetPhysicalAddressHigh(pBufPool->AllocPhysicalAddr) != 0)
    {
        NdisMFreeSharedMemory(pAdapter->hAdapter,
            pBufPool->ulAllocSize,
            pBufPool->bCached,
            pBufPool->pAllocVirtual,
            pBufPool->AllocPhysicalAddr);
        pBufPool->pAllocVirtual = NULL;
        pBufPool->ulAllocSize = 0;
    }
    else
    {
        NdisZeroMemory(pBufPool->pAllocVirtual, pBufPool->ulAllocSize);
        nOffset = (((ULONG_PTR) pBufPool->pAllocVirtual % BUFFER_ALIGNMENT) ?
            (BUFFER_ALIGNMENT - (int)
            ((ULONG_PTR) pBufPool->pAllocVirtual % BUFFER_ALIGNMENT)) : 0);
        pBufPool->pVirtual = pBufPool->pAllocVirtual + nOffset;
        pBufPool->ulPhysical = NdisGetPhysicalAddressLow(
            pBufPool->AllocPhysicalAddr) + nOffset;
        nOffset = 0;
#if DBG
        DbgPrint("buffer: %p %X\n", pBufPool->pVirtual, pBufPool->ulPhysical);
#endif
    }

    pBufInfo->BufArray = AllocMemory(pAdapter->hAdapter,
        sizeof(DMA_BUFFER) * pBufInfo->cnAlloc);
    NdisZeroMemory(pBufInfo->BufArray, sizeof(DMA_BUFFER) * pBufInfo->cnAlloc);

    /* Set up each buffer. */
    pAllocVirtual = NULL;
    NdisSetPhysicalAddressHigh(AllocPhysicalAddr, 0);
    NdisSetPhysicalAddressLow(AllocPhysicalAddr, 0);
    NdisSetPhysicalAddressHigh(PhysicalAddr, 0);
    ulAllocSize = pBufPool->ulAllocSize;

    pDma = pBufInfo->BufArray;
    for (i = 0; i < pBufInfo->cnAlloc; i++)
    {
        /* Big buffer allocated. */
        if (pBufPool->pAllocVirtual)
        {
            /* Allocate each small buffer from big buffer. */
            pVirtual = pBufPool->pVirtual + nOffset;
            ulPhysical = pBufPool->ulPhysical + nOffset;
            NdisSetPhysicalAddressLow(PhysicalAddr, ulPhysical);
            nOffset += NDIS_BUFFER_SIZE;
        }
        else
        {
            ulAllocSize = NDIS_BUFFER_SIZE + BUFFER_ALIGNMENT;

            /* Allocate each small buffer from system. */
            NdisMAllocateSharedMemory(pAdapter->hAdapter,
                ulAllocSize,
                TRUE,
                &pAllocVirtual,
                &AllocPhysicalAddr);
            if (pAllocVirtual == NULL)
            {
                return NDIS_STATUS_RESOURCES;
            }
            else if (NdisGetPhysicalAddressHigh(AllocPhysicalAddr) != 0)
            {
                NdisMFreeSharedMemory(pAdapter->hAdapter,
                    ulAllocSize,
                    TRUE,
                    pAllocVirtual,
                    AllocPhysicalAddr);
                return NDIS_STATUS_RESOURCES;
            }

            NdisZeroMemory(pAllocVirtual, ulAllocSize);
            nOffset = (((ULONG_PTR) pAllocVirtual % BUFFER_ALIGNMENT) ?
                (BUFFER_ALIGNMENT - (int)
                ((ULONG_PTR) pAllocVirtual % BUFFER_ALIGNMENT)) : 0);
            pVirtual = pAllocVirtual + nOffset;
            ulPhysical = NdisGetPhysicalAddressLow(AllocPhysicalAddr) +
                nOffset;
            NdisSetPhysicalAddressLow(PhysicalAddr, ulPhysical);
        }

        /* Save the allocated memory information. */
        pDma->sm.ulAllocSize = ulAllocSize;
        pDma->sm.pAllocVirtual = pAllocVirtual;
        pDma->sm.AllocPhysicalAddr = AllocPhysicalAddr;

        /* These addresses point to the data buffer. */
        pDma->sm.pVirtual = pVirtual;
        pDma->sm.ulPhysical = ulPhysical;
        pDma->PhysicalAddr = PhysicalAddr;
        pDma->ulSize = RX_BUF_SIZE;

        /* Allocate an NDIS flush buffer for each buffer. */
        pDma->Mdl = NdisAllocateMdl(pAdapter->hAdapter, pDma->sm.pVirtual,
            pDma->ulSize);
        if (!pDma->Mdl)
        {
            return NDIS_STATUS_RESOURCES;
        }

        if (fPacket)
        {
            /* Allocate a packet for each receive buffer. */
            pDma->Nbl = NdisAllocateNetBufferAndNetBufferList(
                pAdapter->hNetBufListPool, 0, 0, pDma->Mdl, 0, 0);
            if (!pDma->Nbl)
            {
                return NDIS_STATUS_RESOURCES;
            }
            pDma->Nbl->SourceHandle = pAdapter->hAdapter;
            pDma->pBuffer = NET_BUFFER_LIST_FIRST_NB(pDma->Nbl);
            NBL_FROM_SEND_NB(pDma->pBuffer) = pDma->Nbl;

            /* Save a pointer to the receive buffer with the packet. */
            DMA_FROM_RCV_NBL(pDma->Nbl) = pDma;
        }
        pDma++;
    }
    NdisAllocateSpinLock(&pBufInfo->listLock);
    NdisInitializeSListHead(&pBufInfo->listHead);
    InitBuffers(pBufInfo);

    return NDIS_STATUS_SUCCESS;
}  /* AllocateBuffers */


#ifndef INLINE
/*
    Alloc_DMA_Buffer

Description:
    This function allocates a DMA buffer information block for either
    transmitting or receiving.

Parameters:
    PBUFFER_INFO pBufInfo
        Pointer to DMA buffer information structure.

    PTDesc pCurrent
        Pointer to current descriptor.

Return (PDMA_BUFFER):
    Pointer to the DMA buffer information block if successful; NULL otherwise.
*/

PDMA_BUFFER Alloc_DMA_Buffer (
    PBUFFER_INFO pBufInfo,
    PTDesc       pCurrent )
{
    PDMA_BUFFER        pDma;
    PSINGLE_LIST_ENTRY pEntry;

    pEntry = NdisInterlockedPopEntrySList(&pBufInfo->listHead,
        &pBufInfo->listLock);
    ASSERT(pEntry != NULL);
    pDma = CONTAINING_RECORD(pEntry, DMA_BUFFER, listEntry);

    /* Link to current descriptor. */
    pCurrent->pReserved = pDma;
    return pDma;
}  // Alloc_DMA_Buffer


/*
    Free_DMA_Buffer

Description:
    This routine frees the DMA buffer information block.

Parameters:
    PBUFFER_INFO pBufInfo
        Pointer to DMA buffer information structure.

    PDMA_BUFFER pDma
        Pointer to DMA buffer information block.

Return (None):
*/

VOID Free_DMA_Buffer (
    PBUFFER_INFO pBufInfo,
    PDMA_BUFFER  pDma )
{
    NdisInterlockedPushEntrySList(&pBufInfo->listHead, &pDma->listEntry,
        &pBufInfo->listLock);
}  // Free_DMA_Buffer
#endif


/*
    InitReceiveBuffers

Description:
    This routine initializes DMA buffers for receiving.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

void InitReceiveBuffers (
        PNDIS_ADAPTER pAdapter )
{
    int i;
    PTDesc pDesc;
    PDMA_BUFFER pDma;
    PHARDWARE pHardware = &pAdapter->Hardware;
    PTDescInfo pInfo = &pHardware->m_RxDescInfo;

    for (i = 0; i < pHardware->m_RxDescInfo.cnAlloc; i++)
    {
        GetRxPacket(pInfo, pDesc);

        /* Get a new buffer for this descriptor. */
        pDma = Alloc_DMA_Buffer(&pAdapter->RxBufInfo, pDesc);

        /* Set descriptor. */
        SetReceiveBuffer(pDesc, pDma->sm.ulPhysical);
        SetReceiveLength(pDesc, pDma->ulSize);
        ReleasePacket(pDesc);
    }
}  /* InitReceiveBuffers */


/*
    AllocateMemory

Description:
    This function allocates memory for use by hardware descriptors for
    receiving and transmitting.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

NDIS_STATUS AllocateMemory (
        PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    NDIS_STATUS Status;

    /* Determine the number of receive and transmit descriptors. */

#ifdef SKIP_TX_INT
    pHardware->m_TxIntCnt = 0;
    pHardware->m_TxIntMask = pHardware->m_TxDescInfo.cnAlloc / 4;
    if (pHardware->m_TxIntMask > 8)
    {
        pHardware->m_TxIntMask = 8;
    }
    while (pHardware->m_TxIntMask) {
        pHardware->m_TxIntCnt++;
        pHardware->m_TxIntMask >>= 1;
    }
    if (pHardware->m_TxIntCnt) {
        pHardware->m_TxIntMask = (1 << (pHardware->m_TxIntCnt - 1)) - 1;
        pHardware->m_TxIntCnt = 0;
    }
#endif

    /* Determine the descriptor size. */
    pHardware->m_RxDescInfo.nSize =
        (((sizeof(THw_Desc) + DESC_ALIGNMENT - 1) / DESC_ALIGNMENT) *
        DESC_ALIGNMENT);
    pHardware->m_TxDescInfo.nSize =
        (((sizeof(THw_Desc) + DESC_ALIGNMENT - 1) / DESC_ALIGNMENT) *
        DESC_ALIGNMENT);

    if (pHardware->m_RxDescInfo.nSize != sizeof(THw_Desc))
    {
        DBG_PRINT("Hardware descriptor size not right!"NEWLINE);
    }
    CheckDescriptorNum(&pHardware->m_RxDescInfo);
    CheckDescriptorNum(&pHardware->m_TxDescInfo);

    /* TX/RX buffers we need. */
#if 1
    /* More receive buffers so that they can be used by upper layer and
       returned later in the MiniportReturnNetBufferLists call.
    */
    pAdapter->RxBufInfo.cnAlloc = pHardware->m_RxDescInfo.cnAlloc << 1;
#else
    pAdapter->RxBufInfo.cnAlloc = pHardware->m_RxDescInfo.cnAlloc;
#endif
    pAdapter->TxBufInfo.cnAlloc = pHardware->m_TxDescInfo.cnAlloc;

    /* Allocate map registers. */
    Status = AllocateTcb(pAdapter);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }

    /* Allocate descriptors */
    Status = AllocateDescriptors(pAdapter);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }

    /* Allocate a pool of NDIS buffers. */
    Status = AllocNetBufferListPool(pAdapter);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }

    /* Allocate receive buffers. */
    Status = AllocateBuffers(pAdapter, &pAdapter->RxBufInfo, TRUE);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }

    /* Allocate transmit buffers. */
    Status = AllocateBuffers(pAdapter, &pAdapter->TxBufInfo, FALSE);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        return Status;
    }

    InitReceiveBuffers(pAdapter);

#ifdef DBG_
    CheckDescriptors(&pAdapter->Hardware.m_RxDescInfo);
    CheckDescriptors(&pAdapter->Hardware.m_TxDescInfo);
#endif

    return NDIS_STATUS_SUCCESS;
}  /* AllocateMemory */


/*
    FreeTcb

Description:
    This local routine frees the mapping registers allocated by AllocateTcb().

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

static void FreeTcb (
        PNDIS_ADAPTER pAdapter )
{
    PTCB_INFO pInfo = &pAdapter->TcbInfo;

    if (pInfo->TcbArray)
    {
        NdisFreeMemory(pInfo->TcbArray, sizeof(TCB) * pInfo->cnAlloc, 0);
        pInfo->TcbArray = NULL;
        NdisFreeSpinLock(&pInfo->listLock);
    }

    if (pInfo->ScatterGatherBuffer)
    {
        NdisFreeMemory(pInfo->ScatterGatherBuffer,
            pInfo->ScatterGatherSize * pInfo->cnAlloc, 0);
        pInfo->ScatterGatherBuffer = NULL;
    }

    if (pInfo->hDma)
    {
        NdisMDeregisterScatterGatherDma(pInfo->hDma);
        pInfo->hDma = 0;
    }
}  /* FreeTcb */


/*
    FreeDescriptors

Description:
    This local routine frees the software and hardware descriptors
    allocated by AllocateDescriptors().

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

static void FreeDescriptors (
        PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;

    /* reset descriptor */
    pHardware->m_RxDescInfo.phwRing = NULL;
    pHardware->m_TxDescInfo.phwRing = NULL;
    pHardware->m_RxDescInfo.ulRing = 0;
    pHardware->m_TxDescInfo.ulRing = 0;

    /* Free memory */
    if (pAdapter->DescPool.pAllocVirtual)
        NdisMFreeSharedMemory(pAdapter->hAdapter,
            pAdapter->DescPool.ulAllocSize,
            pAdapter->DescPool.bCached,
            pAdapter->DescPool.pAllocVirtual,
            pAdapter->DescPool.AllocPhysicalAddr);

    /* reset resource pool */
    pAdapter->DescPool.ulAllocSize = 0;
    pAdapter->DescPool.pAllocVirtual = NULL;

    if (pHardware->m_RxDescInfo.pRing)
    {
        NdisFreeMemory(pHardware->m_RxDescInfo.pRing,
            sizeof(TDesc) * pHardware->m_RxDescInfo.cnAlloc, 0);
        pHardware->m_RxDescInfo.pRing = NULL;
    }
    if (pHardware->m_TxDescInfo.pRing)
    {
        NdisFreeMemory(pHardware->m_TxDescInfo.pRing,
            sizeof(TDesc) * pHardware->m_TxDescInfo.cnAlloc, 0);
        pHardware->m_TxDescInfo.pRing = NULL;
    }
}  /* FreeDescriptors */


/*
    FreeBuffers

Description:
    This local routine frees the DMA buffers allocated by
    AllocateBuffers().

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PBUFFER_INFO pBufInfo
        Pointer to DMA buffer information structure.

Return (None):
*/

static void FreeBuffers (
        PNDIS_ADAPTER pAdapter,
        PBUFFER_INFO  pBufInfo )
{
    int i;
    PDMA_BUFFER pDma;
    PSHARED_MEM pBufPool = &pBufInfo->MemPool;

    if (!pBufInfo->BufArray)
        return;

    pDma = pBufInfo->BufArray;
    for (i = 0; i < pBufInfo->cnAlloc; i++)
    {
        if (pDma->Nbl)
        {
            NdisFreeNetBufferList(pDma->Nbl);
            pDma->Nbl = NULL;
        }
        if (pDma->Mdl)
        {
            NdisFreeMdl(pDma->Mdl);
            pDma->Mdl = NULL;
        }
        if (pDma->sm.pAllocVirtual)
        {
            NdisMFreeSharedMemory(pAdapter->hAdapter,
                pDma->sm.ulAllocSize,
                pDma->sm.bCached,
                pDma->sm.pAllocVirtual,
                pDma->sm.AllocPhysicalAddr);
        }
        pDma++;
    }
    NdisFreeMemory(pBufInfo->BufArray,
        sizeof(DMA_BUFFER) * pBufInfo->cnAlloc, 0);
    pBufInfo->BufArray = NULL;

    if (pBufPool->pAllocVirtual)
    {
        NdisMFreeSharedMemory(pAdapter->hAdapter,
            pBufPool->ulAllocSize,
            pBufPool->bCached,
            pBufPool->pAllocVirtual,
            pBufPool->AllocPhysicalAddr);
        pBufPool->ulAllocSize = 0;
        pBufPool->pAllocVirtual = NULL;
    }
    NdisFreeSpinLock(&pBufInfo->listLock);
}  /* FreeBuffers */


/*
    FreeMemory

Description:
    This local routine frees all the resources allocated by AllocateMemory().

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

VOID FreeMemory (
        PNDIS_ADAPTER pAdapter )
{
    /* Free transmit buffers. */
    FreeBuffers(pAdapter, &pAdapter->TxBufInfo);

    /* Free receive buffers */
    FreeBuffers(pAdapter, &pAdapter->RxBufInfo);

    /* Free map registers. */
    FreeTcb(pAdapter);

    /* Free descriptors. */
    FreeDescriptors(pAdapter);

    /* Free buffer list pool handle. */
    if (pAdapter->hNetBufListPool)
    {
        NdisFreeNetBufferListPool(pAdapter->hNetBufListPool);
        pAdapter->hNetBufListPool = 0;
    }
}  /* FreeMemory */

/* -------------------------------------------------------------------------- */

typedef VOID (*SET_ATTRIBUTES)
    (PNDIS_ADAPTER, PNDIS_MINIPORT_ADAPTER_ATTRIBUTES);

static VOID SetGeneralAttributes (
    IN  PNDIS_ADAPTER                     pAdapter,
    OUT PNDIS_MINIPORT_ADAPTER_ATTRIBUTES Attributes )
{
    PNDIS_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES General =
        &Attributes->GeneralAttributes;

    NdisZeroMemory(General, sizeof(NDIS_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES));

#if defined(NDIS620_MINIPORT)
    General->Header.Type = NDIS_OBJECT_TYPE_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES;
    General->Header.Revision =
        NDIS_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES_REVISION_2;
    General->Header.Size =
        NDIS_SIZEOF_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES_REVISION_2;
#else
    General->Header.Type = NDIS_OBJECT_TYPE_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES;
    General->Header.Revision =
        NDIS_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES_REVISION_1;
    General->Header.Size =
        NDIS_SIZEOF_MINIPORT_ADAPTER_GENERAL_ATTRIBUTES_REVISION_1;
#endif

    //
    // Specify the medium type that the NIC can support but not
    // necessarily the medium type that the NIC currently uses.
    //
    General->MediaType = NdisMedium802_3;

    //
    // Specifiy medium type that the NIC currently uses.
    //
    General->PhysicalMediumType = NdisPhysicalMediumUnspecified;

    //
    // Specifiy the maximum network frame size, in bytes, that the NIC
    // supports excluding the header. A NIC driver that emulates another
    // medium type for binding to a transport must ensure that the maximum
    // frame size for a protocol-supplied net buffer does not exceed the
    // size limitations for the true network medium.
    //
    General->MtuSize = NDIS_MAX_LOOKAHEAD;
    General->MaxXmitLinkSpeed = NDIS_MEDIA_MAX_SPEED;
    General->MaxRcvLinkSpeed = NDIS_MEDIA_MAX_SPEED;
    General->XmitLinkSpeed = NDIS_LINK_SPEED_UNKNOWN;
    General->RcvLinkSpeed = NDIS_LINK_SPEED_UNKNOWN;
    General->MediaConnectState = MediaConnectStateUnknown;
    General->MediaDuplexState = MediaDuplexStateUnknown;

    //
    // The maximum number of bytes the NIC can provide as lookahead data.
    // If that value is different from the size of the lookahead buffer
    // supported by bound protocols, NDIS will call MiniportOidRequest to
    // set the size of the lookahead buffer provided by the miniport driver
    // to the minimum of the miniport driver and protocol(s) values. If the
    // driver always indicates up full packets with
    // NdisMIndicateReceiveNetBufferLists, it should set this value to the
    // maximum total frame size, which excludes the header.
    //
    // Upper-layer drivers examine lookahead data to determine whether a
    // packet that is associated with the lookahead data is intended for
    // one or more of their clients. If the underlying driver supports
    // multipacket receive indications, bound protocols are given full net
    // packets on every indication. Consequently, this value is identical
    // to that returned for OID_GEN_RECEIVE_BLOCK_SIZE.
    //
    General->LookaheadSize = NDIS_MAX_LOOKAHEAD;
#ifdef DBG
    DBG_PRINT("max mtu: %u %u\n", General->MtuSize, General->LookaheadSize);
#endif

//
// Specify a bitmask that defines optional properties of the NIC.
// This miniport indicates receive with NdisMIndicateReceiveNetBufferLists
// function.  Such a driver should set this NDIS_MAC_OPTION_TRANSFERS_NOT_PEND
// flag.
//
// NDIS_MAC_OPTION_NO_LOOPBACK tells NDIS that NIC has no internal
// loopback support so NDIS will manage loopbacks on behalf of
// this driver.
//
// NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA tells the protocol that
// our receive buffer is not on a device-specific card. If
// NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA is not set, multi-buffer
// indications are copied to a single flat buffer.
//
    pAdapter->ulMacOptions = NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA |
        NDIS_MAC_OPTION_TRANSFERS_NOT_PEND |
        NDIS_MAC_OPTION_NO_LOOPBACK;
    General->MacOptions = pAdapter->ulMacOptions;

    General->SupportedPacketFilters = NDIS_PACKET_TYPE_DIRECTED |
        NDIS_PACKET_TYPE_PROMISCUOUS |
        NDIS_PACKET_TYPE_MULTICAST |
        NDIS_PACKET_TYPE_ALL_MULTICAST |
        NDIS_PACKET_TYPE_BROADCAST;

    //
    // The maximum number of multicast addresses the NIC driver can manage.
    // This list is global for all protocols bound to (or above) the NIC.
    // Consequently, a protocol can receive NDIS_STATUS_MULTICAST_FULL from
    // the NIC driver when attempting to set the multicast address list,
    // even if the number of elements in the given list is less than the
    // number originally returned for this query.
    //
    General->MaxMulticastListSize = MAX_MULTICAST_LIST;
    General->MacAddressLength = MAC_ADDRESS_LENGTH;

    //
    // Return the MAC address of the NIC burnt in the hardware.
    //
    NdisMoveMemory(General->PermanentMacAddress,
        pAdapter->Hardware.m_bPermanentAddress, MAC_ADDRESS_LENGTH);

    //
    // Return the MAC address the NIC is currently programmed to use. Note
    // that this address could be different from the permananent address as
    // the user can override using registry. Read NdisReadNetworkAddress
    // doc for more info.
    //
    NdisMoveMemory(General->CurrentMacAddress,
        pAdapter->Hardware.m_bOverrideAddress, MAC_ADDRESS_LENGTH);

    General->RecvScaleCapabilities = NULL;
    General->AccessType = NET_IF_ACCESS_BROADCAST;
    General->DirectionType = NET_IF_DIRECTION_SENDRECEIVE;
    General->ConnectionType = NET_IF_CONNECTION_DEDICATED;
    General->IfType = IF_TYPE_ETHERNET_CSMACD;
    General->IfConnectorPresent = TRUE;

    pAdapter->ulStatistics = NDIS_STATISTICS_XMIT_OK_SUPPORTED |
        NDIS_STATISTICS_RCV_OK_SUPPORTED |
        NDIS_STATISTICS_XMIT_ERROR_SUPPORTED |
        NDIS_STATISTICS_RCV_ERROR_SUPPORTED |
        NDIS_STATISTICS_RCV_CRC_ERROR_SUPPORTED |
        NDIS_STATISTICS_RCV_NO_BUFFER_SUPPORTED |
        NDIS_STATISTICS_TRANSMIT_QUEUE_LENGTH_SUPPORTED |
        NDIS_STATISTICS_GEN_STATISTICS_SUPPORTED;
    General->SupportedStatistics = pAdapter->ulStatistics;

    //
    // The SupportedOidList is an array of OIDs for objects that the
    // underlying driver or its NIC supports.  Objects include general,
    // media-specific, and implementation-specific objects. NDIS forwards a
    // subset of the returned list to protocols that make this query. That
    // is, NDIS filters any supported statistics OIDs out of the list
    // because protocols never make statistics queries.
    //
    General->SupportedOidList = SupportedOids;
    pAdapter->OidsSize = sizeof(SupportedOids);
    General->SupportedOidListLength = pAdapter->OidsSize;

    //
    // Set the power management capabilities.  The format used is NDIS
    // version-specific.
    //
#if defined(NDIS620_MINIPORT)
    pAdapter->PmCapabilities.Header.Type = NDIS_OBJECT_TYPE_DEFAULT;
    pAdapter->PmCapabilities.Header.Size =
        NDIS_SIZEOF_NDIS_PM_CAPABILITIES_REVISION_1;
    pAdapter->PmCapabilities.Header.Revision = NDIS_PM_CAPABILITIES_REVISION_1;

    if (pAdapter->PmCapabilities.MinPatternWakeUp !=
            NdisDeviceStateUnspecified)
    {
        pAdapter->PmCapabilities.SupportedWoLPacketPatterns =
            NDIS_PM_WOL_BITMAP_PATTERN_SUPPORTED |
            NDIS_PM_WOL_MAGIC_PACKET_SUPPORTED;
        pAdapter->PmCapabilities.NumTotalWoLPatterns = 4 + 0;
        pAdapter->PmCapabilities.MaxWoLPatternSize = 64;
        pAdapter->PmCapabilities.MaxWoLPatternOffset = 64 - 0;
        pAdapter->PmCapabilities.MaxWoLPacketSaveBuffer = 0;
    }
#ifdef DBG
    DBG_PRINT("Pm: %d %d %d\n",
        pAdapter->PmCapabilities.MinMagicPacketWakeUp,
        pAdapter->PmCapabilities.MinPatternWakeUp,
        pAdapter->PmCapabilities.MinLinkChangeWakeUp);
#endif

    General->PowerManagementCapabilitiesEx = &pAdapter->PmCapabilities;
#else
#ifdef DBG
    DBG_PRINT("PnP: %d %d %d\n",
        pAdapter->PnpCapabilities.WakeUpCapabilities.MinMagicPacketWakeUp,
        pAdapter->PnpCapabilities.WakeUpCapabilities.MinPatternWakeUp,
        pAdapter->PnpCapabilities.WakeUpCapabilities.MinLinkChangeWakeUp);
#endif

    General->PowerManagementCapabilities = &pAdapter->PnpCapabilities;
#endif
}  // SetGeneralAttributes


static VOID SetRegistrationAttributes (
    IN  PNDIS_ADAPTER                     pAdapter,
    OUT PNDIS_MINIPORT_ADAPTER_ATTRIBUTES Attributes )
{
    PNDIS_MINIPORT_ADAPTER_REGISTRATION_ATTRIBUTES Registration =
        &Attributes->RegistrationAttributes;

    NdisZeroMemory(Registration,
        sizeof(NDIS_MINIPORT_ADAPTER_REGISTRATION_ATTRIBUTES));

    Registration->Header.Type =
        NDIS_OBJECT_TYPE_MINIPORT_ADAPTER_REGISTRATION_ATTRIBUTES;
    Registration->Header.Revision =
        NDIS_MINIPORT_ADAPTER_REGISTRATION_ATTRIBUTES_REVISION_1;
    Registration->Header.Size =
        NDIS_SIZEOF_MINIPORT_ADAPTER_REGISTRATION_ATTRIBUTES_REVISION_1;

    Registration->MiniportAdapterContext = (NDIS_HANDLE) pAdapter;
    Registration->AttributeFlags =
        NDIS_MINIPORT_ATTRIBUTES_HARDWARE_DEVICE |
#if 0
        NDIS_MINIPORT_ATTRIBUTES_SURPRISE_REMOVE_OK |
#endif
        NDIS_MINIPORT_ATTRIBUTES_BUS_MASTER;

    Registration->CheckForHangTimeInSeconds = 2;
    Registration->InterfaceType = NdisInterfacePci;
}  // SetRegistrationAttributes


static
NDIS_STATUS SetMiniportAttributes (
    IN  PNDIS_ADAPTER  pAdapter,
    IN  SET_ATTRIBUTES SetAttributes )
{
    NDIS_MINIPORT_ADAPTER_ATTRIBUTES Attributes;

    SetAttributes(pAdapter, &Attributes);
    return NdisMSetMiniportAttributes(pAdapter->hAdapter, &Attributes);
}  // SetMiniportAttributes


/*
    InitAdapter

Description:
    This function initializes the adapter when the adapter is started.  It
    calls ReadConfiguration() to get all information required to start the
    adapter.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNDIS_MINIPORT_INIT_PARAMETERS InitParameters
        Init parameters from driver.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

NDIS_STATUS InitAdapter (
        PNDIS_ADAPTER pAdapter,
    IN  PNDIS_MINIPORT_INIT_PARAMETERS InitParameters )
{
    NDIS_STATUS Status;

    Status = ReadConfiguration(pAdapter, InitParameters);
    if (!NT_SUCCESS(Status))
    {
        return Status;
    }

    /* Need to set bus master else NdisMAllocateSharedMemory will not succeed.
    */
    Status = SetMiniportAttributes(pAdapter, SetRegistrationAttributes);
    if (!NT_SUCCESS(AllocateMemory(pAdapter)))
    {
        return NDIS_STATUS_FAILURE;
    }
    NdisInitializeListHead(&pAdapter->SendWaitList);
    NdisAllocateSpinLock(&pAdapter->SendWaitListLock);

    return NDIS_STATUS_SUCCESS;
}  // InitAdapter


/*
    SetupAdapter

Description:
    This function setups the adapter when the adapter is started.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

NDIS_STATUS SetupAdapter (
    PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    UCHAR bPromiscuous;

    if (!AcquireAdapter(pAdapter, FALSE))
    {
        return NDIS_STATUS_ADAPTER_NOT_FOUND;
    }
    if (!HardwareInitialize(pHardware) || !HardwareReset(pHardware))
    {
        NdisWriteErrorLogEntry(pAdapter->hAdapter,
            NDIS_ERROR_CODE_HARDWARE_FAILURE, 0);
        ReleaseAdapter(pAdapter);
        return NDIS_STATUS_ADAPTER_NOT_FOUND;
    }

    HardwareDisable(pHardware);
    HardwareSetPowerSaving(pHardware, FALSE);

#if 0
    HardwareGetVersion( pHardware );
#endif

    // Read the Ethernet address.
    if (!HardwareReadAddress(pHardware))
    {
        NdisWriteErrorLogEntry(pAdapter->hAdapter,
            NDIS_ERROR_CODE_HARDWARE_FAILURE, 0);
        ReleaseAdapter(pAdapter);
        return NDIS_STATUS_ADAPTER_NOT_FOUND;
    }

    pHardware->m_ulHardwareState = MediaStateDisconnected;

    /* Initialize to invalid value so that link detection is done. */
    pHardware->m_PortInfo[MAIN_PORT].bLinkPartner = 0xFF;
    pHardware->m_PortInfo[MAIN_PORT].ulHardwareState =
        MediaStateDisconnected;

#ifdef DEF_KS8842
    pHardware->m_PortInfo[OTHER_PORT].bLinkPartner = 0xFF;
    pHardware->m_PortInfo[OTHER_PORT].ulHardwareState =
        MediaStateDisconnected;
#endif

    HardwareSetup(pHardware);
    if (!pAdapter->ulRxChecksum)
    {
        pHardware->m_dwReceiveConfig &= ~(DMA_RX_CTRL_CSUM_TCP |
            DMA_RX_CTRL_CSUM_UDP | DMA_RX_CTRL_CSUM_IP);
    }
    HardwareSwitchSetup(pHardware);
    HardwareSetDescriptorBase(pHardware,
        pHardware->m_TxDescInfo.ulRing,
        pHardware->m_RxDescInfo.ulRing);

    if ((pAdapter->ulPacketFilter &
            (NDIS_PACKET_TYPE_PROMISCUOUS | NDIS_PACKET_TYPE_ALL_LOCAL)))
        bPromiscuous = TRUE;
    else
        bPromiscuous = FALSE;
    if (!HardwareSetPromiscuous(pHardware, bPromiscuous))
    {
        ReleaseAdapter(pAdapter);
        return NDIS_STATUS_ADAPTER_NOT_FOUND;
    }

    pAdapter->ulNdisMediaState = pAdapter->Hardware.m_ulHardwareState;

    PortInitCounters(pHardware, MAIN_PORT);

#if defined(DEF_KS8842) && !defined(TWO_NETWORK_INTERFACE)
    PortInitCounters(pHardware, OTHER_PORT);
#endif
    PortInitCounters(pHardware, HOST_PORT);

    HardwareSetupInterrupt(pHardware);

#ifdef DBG
DbgPrint("Setup completed"NEWLINE);
#endif

#if 0
    HardwareGetInfo( pHardware );
#endif

    ReleaseAdapter(pAdapter);

#if defined(NDIS620_MINIPORT)
    pAdapter->PmCapabilities.MinMagicPacketWakeUp =
    pAdapter->PmCapabilities.MinPatternWakeUp =
    pAdapter->PmCapabilities.MinLinkChangeWakeUp =
#else
    pAdapter->PnpCapabilities.WakeUpCapabilities.MinMagicPacketWakeUp =
    pAdapter->PnpCapabilities.WakeUpCapabilities.MinPatternWakeUp =
    pAdapter->PnpCapabilities.WakeUpCapabilities.MinLinkChangeWakeUp =
#endif
        NdisDeviceStateUnspecified;

#ifdef DEF_KS8841
    /* Specifies the lowest device power state from which the
       miniport's NIC can signal a wake-up on receipt of a Magic
       Packet.  (A Magic Packet is a packet that contains 16 contiguous
       copies of the receiving NIC's Ethernet address.)
    */
#if defined(NDIS620_MINIPORT)
    pAdapter->PmCapabilities.MinMagicPacketWakeUp =
#else
    pAdapter->PnpCapabilities.WakeUpCapabilities.MinMagicPacketWakeUp =
#endif
        NdisDeviceStateD3;

    /* Specifies the lowest device power state from which the
       miniport's NIC can signal a wake-up event on receipt of a
       network frame that contains a pattern specified by the protocol
       driver.
    */
#if defined(NDIS620_MINIPORT)
    pAdapter->PmCapabilities.MinPatternWakeUp =
#else
    pAdapter->PnpCapabilities.WakeUpCapabilities.MinPatternWakeUp =
#endif
        NdisDeviceStateD3;

    /* Specifies the lowest device power state from which the
       miniport's NIC can signal a wake-up event in response to a link
       change (the connection or disconnection of the NIC's network
       cable).
    */
#if defined(NDIS620_MINIPORT)
    pAdapter->PmCapabilities.MinLinkChangeWakeUp =
#else
    pAdapter->PnpCapabilities.WakeUpCapabilities.MinLinkChangeWakeUp =
#endif
        NdisDeviceStateD3;
#endif

    pAdapter->CurrentPowerState = pAdapter->NextPowerState =
        NdisDeviceStateD0;

    SetMiniportAttributes(pAdapter, SetGeneralAttributes);

    return NDIS_STATUS_SUCCESS;
}  // SetupAdapter


/*
    ShutdownAdapter

Description:
    This routine shutdowns the adapter when the driver is unloaded.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

void ShutdownAdapter (
    PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
#ifdef DBG_
    int i;
#endif

    AcquireAdapter(pAdapter, FALSE);
    HardwareDisableInterrupt(pHardware);
    HardwareDisable(pHardware);
    ReleaseAdapter(pAdapter);

#ifdef DBG_
    DbgPrint("counters:\n");
    for (i = OID_COUNTER_FIRST; i < OID_COUNTER_LAST; i++)
    {
        DbgPrint("%u = %u\n", i, pHardware->m_cnCounter[0][i]);
    }
    DbgPrint("wait delays:\n");
    for (i = WAIT_DELAY_FIRST; i < WAIT_DELAY_LAST; i++)
    {
        DbgPrint("%u = %u\n", i, pHardware->m_nWaitDelay[i]);
    }
    DbgPrint("bad:\n");
    for (i = COUNT_BAD_FIRST; i < COUNT_BAD_LAST; i++)
    {
        DbgPrint("%u = %u\n", i, pHardware->m_nBad[i]);
    }
    DbgPrint("good:\n");
    for (i = COUNT_GOOD_FIRST; i < COUNT_GOOD_LAST; i++)
    {
        DbgPrint("%u = %u\n", i, pHardware->m_nGood[i]);
    }
#endif
}  // ShutdownAdapter


/*
    StartAdapter

Description:
    This routine starts the adapter when the driver is woke up.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

void StartAdapter (
    PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;

    AcquireAdapter(pAdapter, FALSE);
    HardwareEnable(pHardware);
    HardwareEnableInterrupt(pHardware);
    ReleaseAdapter(pAdapter);
}  // StartAdapter

/* -------------------------------------------------------------------------- */

#define INTR_ACQ_TIMEOUT  300

VOID AcquireInterrupt (
    PNDIS_ADAPTER pAdapter,
    int           operate )
{
    int timeout = INTR_ACQ_TIMEOUT;
#ifdef DEBUG_INTR_LOCK
    int operating = 0;

    while (1 != InterlockedIncrement(&pAdapter->OperateAcquire))
        InterlockedDecrement(&pAdapter->OperateAcquire);
    pAdapter->InterruptOperate |= operate << 16;
    InterlockedDecrement(&pAdapter->OperateAcquire);
#endif
    while (1 != InterlockedIncrement(&pAdapter->InterruptAcquire))
    {
#ifdef DEBUG_INTR_LOCK
        if (!operating)
        {
            while (1 != InterlockedIncrement(&pAdapter->OperateAcquire))
                InterlockedDecrement(&pAdapter->OperateAcquire);
            operating = pAdapter->InterruptOperate;
            InterlockedDecrement(&pAdapter->OperateAcquire);
        }
#endif
        NdisInterlockedDecrement(&pAdapter->InterruptAcquire);
        if (timeout)
            --timeout;
        else
        {
            NdisInterlockedIncrement(&pAdapter->InterruptAcquire);
#ifdef DBG
            DBG_PRINT(" intr acq timeout\n");
#endif
            break;
        }
    }
#ifdef DEBUG_INTR_LOCK
    while (1 != InterlockedIncrement(&pAdapter->OperateAcquire))
        InterlockedDecrement(&pAdapter->OperateAcquire);
    pAdapter->InterruptOperate |= operate;
    InterlockedDecrement(&pAdapter->OperateAcquire);
#if 0
    if (timeout < INTR_ACQ_TIMEOUT)
        DBG_PRINT("acqintr: %d; %x, %x\n", INTR_ACQ_TIMEOUT - timeout,
            operate, operating);
#endif
#endif
}  // AcquireInterrupt


VOID ReleaseInterrupt (
    PNDIS_ADAPTER pAdapter,
    int           operate )
{
#ifdef DEBUG_INTR_LOCK
    int operating;

    operate |= operate << 16;
    while (1 != InterlockedIncrement(&pAdapter->OperateAcquire))
        InterlockedDecrement(&pAdapter->OperateAcquire);
    operating = pAdapter->InterruptOperate;
    pAdapter->InterruptOperate &= ~operate;
    InterlockedDecrement(&pAdapter->OperateAcquire);
    if ((operating & operate) != operate)
        DBG_PRINT(" relintr: %x %x\n", operate, operating);
#endif
    NdisInterlockedDecrement(&pAdapter->InterruptAcquire);
}  // ReleaseInterrupt


BOOLEAN HardwareBlockIntrSynchronize (
    IN  PVOID Context )
{
    PHARDWARE pHardware = Context;
    PNDIS_ADAPTER pAdapter = pHardware->m_pDevice;

    AcquireInterrupt(pAdapter, INTR_ACQ_BLOCK);
    HardwareBlockInterrupt_PCI(pHardware);
    ReleaseInterrupt(pAdapter, INTR_ACQ_BLOCK);
    return TRUE;
}  // HardwareBlockIntrSynchronize


BOOLEAN HardwareRestoreIntrSynchronize (
    IN  PVOID Context )
{
    PHARDWARE pHardware = Context;
    PNDIS_ADAPTER pAdapter = pHardware->m_pDevice;

    AcquireInterrupt(pAdapter, INTR_ACQ_RESTORE);
    HardwareRestoreInterrupt_PCI(pHardware, pHardware->m_ulInterruptBlocked);
    ReleaseInterrupt(pAdapter, INTR_ACQ_RESTORE);
    return TRUE;
}  // HardwareRestoreIntrSynchronize


UINT HardwareBlockIntrSync (
    void* pContext )
{
    PHARDWARE pHardware = pContext;
    UINT Interrupt;
    PNDIS_ADAPTER pAdapter = pHardware->m_pDevice;

    NdisMSynchronizeWithInterruptEx(pAdapter->hInterrupt, 0,
        HardwareBlockIntrSynchronize, pHardware);
    Interrupt = pHardware->m_ulInterruptBlocked;
    return Interrupt;
}  // HardwareBlockIntrSync


VOID HardwareRestoreIntrSync (
    void* pContext,
    UINT  Interrupt )
{
    PHARDWARE pHardware = pContext;
    PNDIS_ADAPTER pAdapter = pHardware->m_pDevice;

    NdisMSynchronizeWithInterruptEx(pAdapter->hInterrupt, 0,
        HardwareRestoreIntrSynchronize, pHardware);
}  // HardwareRestoreIntrSync


/*
    MiniportSynchronizeISR

Description:
    This function is used to do things not interruptable by software
    interrupt.

Parameters:
    PVOID Context
        Pointer to hardware information structure.

Return (BOOLEAN):
    TRUE if successful; otherwise, FALSE.
*/

BOOLEAN MiniportSynchronizeISR (
    IN  PVOID Context )
{
    PHARDWARE pHardware = (PHARDWARE) Context;

    HardwareDisableInterrupt(pHardware);
    return TRUE;
}  // MiniportSynchronizeISR


/*
    HardwareDisableInterruptSync

Description:
    This function is used to disable hardware interrupt atomically so that
    the operation is not interrupted.

Parameters:
    void* pHardware
        Pointer to hardware information structure.

Return (None):
*/

void HardwareDisableInterruptSync (
    void* pHardware )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER)((PHARDWARE) pHardware)->
        m_pDevice;

    NdisMSynchronizeWithInterruptEx(pAdapter->hInterrupt, 0,
        MiniportSynchronizeISR, pHardware);
}  /* HardwareDisableInterruptSync */

/* -------------------------------------------------------------------------- */

#define ADAPTER_LOCK_TIMEOUT  100

/*
    AcquireAdapter

Description:
    This function is used to acquire the adapter so that only one process
    has access to the adapter.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    BOOLEAN fWait
        Indicate whether to wait or not.  User functions should wait, while
        periodic timer routine should not.

Return (BOOLEAN):
    TRUE if successful; otherwise, FALSE.
*/

BOOLEAN AcquireAdapter (
    IN  PNDIS_ADAPTER pAdapter,
    IN  BOOLEAN       fWait )
{
    BOOLEAN wait;
    PHARDWARE pHardware = &pAdapter->Hardware;
    int cnTimeOut = ADAPTER_LOCK_TIMEOUT;

    while (--cnTimeOut)
    {
        NdisAcquireSpinLock(&pAdapter->lockHardware);
        if (!pHardware->m_bAcquire)
        {
            pHardware->m_bAcquire = TRUE;
            NdisReleaseSpinLock(&pAdapter->lockHardware);
            return TRUE;
        }
        NdisReleaseSpinLock(&pAdapter->lockHardware);
        if (!fWait)
            break;
        NdisResetEvent(&pAdapter->eventHardware);
        wait = NdisWaitEvent(&pAdapter->eventHardware, 10);
    }
    return FALSE;
}  // AcquireAdapter


/*
    ReleaseAdapter

Description:
    This routine is used to release the adapater so that other process can
    have access to the adapter.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

void ReleaseAdapter (
    IN  PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    LONG acquire;

    NdisAcquireSpinLock(&pAdapter->lockHardware);
    pHardware->m_bAcquire = FALSE;
    NdisReleaseSpinLock(&pAdapter->lockHardware);
    NdisSetEvent(&pAdapter->eventHardware);
}  // ReleaseAdapter

/* -------------------------------------------------------------------------- */

/*
    AdapterCheckForHang

Description:
    This function checks the adapter is working properly.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (BOOLEAN):
    TRUE if adapter is not working; FALSE if working properly.
*/

BOOLEAN AdapterCheckForHang (
    IN  PNDIS_ADAPTER pAdapter )
{
    BOOLEAN hanged = (pAdapter->ulDriverState & DRIVER_STATE_HANGED) ?
        TRUE : FALSE;
#if SET_DEFAULT_LED
    PHARDWARE pHardware = &pAdapter->Hardware;
    USHORT    wData;
#endif

    if (!AcquireAdapter(pAdapter, TRUE))
        return FALSE;

#if SET_DEFAULT_LED
    HW_READ_WORD(pHardware, REG_SWITCH_CTRL_5_OFFSET, &wData);
    if ((TO_LO_BYTE(wData) & SET_DEFAULT_LED) != SET_DEFAULT_LED)
    {
        hanged = TRUE;
    }
#endif

#ifdef DEBUG_INTR_LOCK
    pAdapter->InterruptOperate |= INTR_ACQ_CHECK << 16;
    if (1 == NdisInterlockedIncrement(&pAdapter->InterruptAcquire))
    {
        PHARDWARE pHardware = &pAdapter->Hardware;
        ULONG Data;

        pAdapter->InterruptOperate |= INTR_ACQ_CHECK;
        HW_READ_DWORD(pHardware, REG_INTERRUPTS_ENABLE, &Data);
        if (!Data && !pHardware->m_nBlock)
        {
             DBG_PRINT("  hanged! %x %x\n",
                 pHardware->m_ulInterruptBlocked, pHardware->m_ulInterruptSet);
        }
    }
    pAdapter->InterruptOperate &= ~(INTR_ACQ_CHECK | (INTR_ACQ_CHECK << 16));
    NdisInterlockedDecrement(&pAdapter->InterruptAcquire);
#endif
    if (pAdapter->nNotSend && pAdapter->nBusySend >= pAdapter->nNotSend) {
#ifdef DBG
        UINT enable;
        UINT status;

        HW_READ_DWORD(&pAdapter->Hardware, REG_INTERRUPTS_ENABLE, &enable);
        HW_READ_DWORD(&pAdapter->Hardware, REG_INTERRUPTS_STATUS, &status);
        DbgPrint("%x %x; %d\n", enable, status, pAdapter->nBusySend);
#endif
        if (++pAdapter->nHangCnt >= 2) {
            pAdapter->nHangCnt = 0;
            hanged = TRUE;
        }
    } else
        pAdapter->nHangCnt = 0;
    pAdapter->nNotSend = pAdapter->nBusySend;

    ReleaseAdapter(pAdapter);

    return hanged;
}  // AdapterCheckForHang


/*
    AdapterGetDescription

Description:
    This routine returns the adapter description.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

VOID AdapterGetDescription (
    PVOID* ppSource,
    PUINT  puiLength )
{
#ifdef DEF_KS8841
    static char gszDescription[] = "KSZ8841 PCI Ethernet";
#else
    static char gszDescription[] = "KSZ8842 PCI Ethernet";
#endif

    *ppSource = gszDescription;
    *puiLength = strlen(gszDescription);
}  // AdapterGetDescription


/*
    AdapterReadCounter

Description:
    This function retrieves the statistics counter value.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    ULONG OidCounter
        The Oid counter index.

Return (ULONGLONG):
    The counter value.
*/

ULONGLONG AdapterReadCounter (
    PNDIS_ADAPTER pAdapter,
    ULONG         OidCounter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    PPORT_CONFIG pPort = &pHardware->m_Port[MAIN_PORT];
    ULONGLONG qHardware;
    ULONGLONG qSoftware;
    ULONGLONG qCounter = 0;

    switch (OidCounter)
    {
        case OID_GEN_XMIT_OK:
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_BROADCAST] +
                pPort->cnCounter[MIB_COUNTER_RX_MULTICAST] +
                pPort->cnCounter[MIB_COUNTER_RX_UNICAST];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_BROADCAST] +
                pPort->cnCounter[MIB_COUNTER_TX_MULTICAST] +
                pPort->cnCounter[MIB_COUNTER_TX_UNICAST];
#endif
            qSoftware = pHardware->m_cnCounter[MAIN_PORT][OID_COUNTER_XMIT_OK];
#ifdef DBG
            if (pAdapter->bCheckCounter && qSoftware != qHardware)
                DbgPrint("tx %llu %llu\n", qHardware, qSoftware);
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_RCV_OK:
            qCounter = pHardware->m_cnCounter[MAIN_PORT][OID_COUNTER_RCV_OK];
            break;
        case OID_GEN_XMIT_ERROR:
            qCounter = 0;
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_SINGLE_COLLISION];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_MULTI_COLLISION];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_DEFERRED];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_TOTAL_COLLISION];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_LATE_COLLISION];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_SINGLE_COLLISION];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_MULTI_COLLISION];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_DEFERRED];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_TOTAL_COLLISION];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_LATE_COLLISION];
#endif
            break;
        case OID_GEN_RCV_ERROR:
            qCounter = 0;
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_UNDERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_FRAGMENT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_OVERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_JABBER];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_SYMBOL_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_CRC_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_ALIGNMENT_ERR];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_UNDERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_FRAGMENT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_OVERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_JABBER];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_SYMBOL_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_CRC_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_ALIGNMENT_ERR];
#endif
            break;

        case OID_GEN_BYTES_RCV:
            qSoftware = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_DIRECTED_BYTES_RCV] +
                pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_BROADCAST_BYTES_RCV] +
                pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_MULTICAST_BYTES_RCV];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_LO_PRIORITY] +
                pPort->cnCounter[MIB_COUNTER_TX_HI_PRIORITY];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_LO_PRIORITY] +
                pPort->cnCounter[MIB_COUNTER_RX_HI_PRIORITY];
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_BYTES_XMIT:
            qSoftware = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_DIRECTED_BYTES_XMIT] +
                pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_BROADCAST_BYTES_XMIT] +
                pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_MULTICAST_BYTES_XMIT];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_LO_PRIORITY] +
                pPort->cnCounter[MIB_COUNTER_RX_HI_PRIORITY];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_LO_PRIORITY] +
                pPort->cnCounter[MIB_COUNTER_TX_HI_PRIORITY];
#endif
#ifdef DBG
            if (pAdapter->bCheckCounter && qSoftware != qHardware)
                DbgPrint("tx bytes %llu %llu\n", qHardware, qSoftware);
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_RCV_DISCARDS:
            qCounter = 0;
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_UNDERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_FRAGMENT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_OVERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_JABBER];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_SYMBOL_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_CRC_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_ALIGNMENT_ERR];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_UNDERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_FRAGMENT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_OVERSIZE];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_JABBER];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_SYMBOL_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_CRC_ERR];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_ALIGNMENT_ERR];
#endif
            qCounter += pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_RCV_NO_BUFFER];
            break;
        case OID_GEN_XMIT_DISCARDS:
            break;

        case OID_GEN_RCV_NO_BUFFER:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_RCV_NO_BUFFER];
            break;
        case OID_802_3_RCV_OVERRUN:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_RCV_NO_BUFFER];
            break;

        case OID_GEN_DIRECTED_FRAMES_XMIT:
            qSoftware = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_DIRECTED_FRAME_XMIT];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_UNICAST];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_UNICAST];
#endif
#ifdef DBG
            if (pAdapter->bCheckCounter && qSoftware != qHardware)
                DbgPrint("tx ucast %llu %llu\n", qHardware, qSoftware);
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_BROADCAST_FRAMES_XMIT:
            qSoftware = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_BROADCAST_FRAME_XMIT];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_BROADCAST];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_BROADCAST];
#endif
#ifdef DBG
            if (pAdapter->bCheckCounter && qSoftware != qHardware)
                DbgPrint("tx bcast %llu %llu\n", qHardware, qSoftware);
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_MULTICAST_FRAMES_XMIT:
            qSoftware = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_MULTICAST_FRAME_XMIT];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_MULTICAST];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_MULTICAST];
#endif
#ifdef DBG
            if (pAdapter->bCheckCounter && qSoftware != qHardware)
                DbgPrint("tx mcast %llu %llu\n", qHardware, qSoftware);
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_DIRECTED_BYTES_XMIT:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_DIRECTED_BYTES_XMIT];
            break;
        case OID_GEN_BROADCAST_BYTES_XMIT:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_BROADCAST_BYTES_XMIT];
            break;
        case OID_GEN_MULTICAST_BYTES_XMIT:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_MULTICAST_BYTES_XMIT];
            break;

        case OID_GEN_DIRECTED_FRAMES_RCV:
            qSoftware = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_DIRECTED_FRAME_RCV];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_UNICAST];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_UNICAST];
#endif
#ifdef DBG
            if (pAdapter->bCheckCounter && qSoftware != qHardware)
                DbgPrint("rx ucast %llu %llu\n", qHardware, qSoftware);
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_BROADCAST_FRAMES_RCV:
            qSoftware = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_BROADCAST_FRAME_RCV];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[HOST_PORT];
            qHardware = pPort->cnCounter[MIB_COUNTER_TX_BROADCAST];
#else
            qHardware = pPort->cnCounter[MIB_COUNTER_RX_BROADCAST];
#endif
#ifdef DBG
            if (pAdapter->bCheckCounter && qSoftware != qHardware)
                DbgPrint("rx bcast %llu %llu\n", qHardware, qSoftware);
#endif
            qCounter = qSoftware;
            break;
        case OID_GEN_MULTICAST_FRAMES_RCV:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_MULTICAST_FRAME_RCV];
            break;
        case OID_GEN_DIRECTED_BYTES_RCV:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_DIRECTED_BYTES_RCV];
            break;
        case OID_GEN_BROADCAST_BYTES_RCV:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_BROADCAST_BYTES_RCV];
            break;
        case OID_GEN_MULTICAST_BYTES_RCV:
            qCounter = pHardware->m_cnCounter[MAIN_PORT]
                [OID_COUNTER_MULTICAST_BYTES_RCV];
            break;

        case OID_GEN_RCV_CRC_ERROR:
            qCounter = pPort->cnCounter[MIB_COUNTER_RX_CRC_ERR];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_CRC_ERR];
#endif
            break;
        case OID_802_3_RCV_ERROR_ALIGNMENT:
            qCounter = pPort->cnCounter[MIB_COUNTER_RX_ALIGNMENT_ERR];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_RX_ALIGNMENT_ERR];
#endif
            break;

        case OID_802_3_XMIT_ONE_COLLISION:
            qCounter = pPort->cnCounter[MIB_COUNTER_TX_SINGLE_COLLISION];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_SINGLE_COLLISION];
#endif
            break;
        case OID_802_3_XMIT_MORE_COLLISIONS:
            qCounter = pPort->cnCounter[MIB_COUNTER_TX_MULTI_COLLISION];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_MULTI_COLLISION];
#endif
            break;
        case OID_802_3_XMIT_DEFERRED:
            qCounter = pPort->cnCounter[MIB_COUNTER_TX_DEFERRED];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_DEFERRED];
#endif
            break;
        case OID_802_3_XMIT_MAX_COLLISIONS:
            qCounter = pPort->cnCounter[MIB_COUNTER_TX_TOTAL_COLLISION];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_TOTAL_COLLISION];
#endif
            break;
        case OID_802_3_XMIT_LATE_COLLISIONS:
            qCounter = pPort->cnCounter[MIB_COUNTER_TX_LATE_COLLISION];
#if defined(DEF_KS8842)
            pPort = &pHardware->m_Port[OTHER_PORT];
            qCounter += pPort->cnCounter[MIB_COUNTER_TX_LATE_COLLISION];
#endif
            break;
        default:
#ifdef DBG
            DBG_PRINT("missed: %x\n", OidCounter);
#endif
            break;
    }

    return qCounter;
}  // AdapterReadCounter


/*
    AdapterReset

Description:
    This function resets the adapter.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (BOOLEAN):
    TRUE if successful; otherwise, FALSE.
*/

BOOLEAN AdapterReset (
    PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;

    if (!AcquireAdapter(pAdapter, FALSE))
        return FALSE;

    if (NdisDeviceStateD0 == pAdapter->CurrentPowerState)
    {
        HardwareDisableInterrupt(pHardware);
        HardwareDisable(pHardware);

        HardwareReset(pHardware);
    }
    {
        pHardware->m_ulHardwareState = MediaStateDisconnected;

        /* Initialize to invalid value so that link detection is done. */
        pHardware->m_PortInfo[MAIN_PORT].bLinkPartner = 0xFF;
        pHardware->m_PortInfo[MAIN_PORT].ulHardwareState =
            MediaStateDisconnected;

#ifdef DEF_KS8842
        pHardware->m_PortInfo[OTHER_PORT].bLinkPartner = 0xFF;
        pHardware->m_PortInfo[OTHER_PORT].ulHardwareState =
            MediaStateDisconnected;
#endif

        HardwareSetup(pHardware);
        if (!pAdapter->ulRxChecksum)
        {
            pHardware->m_dwReceiveConfig &= ~(DMA_RX_CTRL_CSUM_TCP |
                DMA_RX_CTRL_CSUM_UDP | DMA_RX_CTRL_CSUM_IP);
        }
        HardwareSwitchSetup(pHardware);
        HardwareSetDescriptorBase(pHardware,
            pHardware->m_TxDescInfo.ulRing,
            pHardware->m_RxDescInfo.ulRing);

        /* Reset may wipe out these registers. */
        if (pHardware->m_bMacOverrideAddr)
            HardwareSetAddress(pHardware);
        if (pHardware->m_bMulticastListSize)
            HardwareSetGroupAddress(pHardware);

        TransmitReset(pAdapter);
        HardwareResetRxPackets(pHardware);
        HardwareResetTxPackets(pHardware);

        NdisInterlockedFlushSList(&pAdapter->RxBufInfo.listHead);
        NdisInterlockedFlushSList(&pAdapter->TxBufInfo.listHead);
        InitBuffers(&pAdapter->RxBufInfo);
        InitBuffers(&pAdapter->TxBufInfo);
        NdisInterlockedFlushSList(&pAdapter->TcbInfo.listHead);
        InitTcb(pAdapter);
        InitReceiveBuffers(pAdapter);

        memset((void *) pHardware->m_cnCounter[MAIN_PORT], 0,
            (sizeof(ULONGLONG) * OID_COUNTER_LAST));
        PortInitCounters(pHardware, MAIN_PORT);

#if defined(DEF_KS8842) && !defined(TWO_NETWORK_INTERFACE)
        PortInitCounters(pHardware, OTHER_PORT);
#endif
        PortInitCounters(pHardware, HOST_PORT);

        HardwareEnable(pHardware);
        HardwareEnableInterrupt(pHardware);
        ReleaseAdapter(pAdapter);
        return TRUE;
    }
    ReleaseAdapter(pAdapter);
    return FALSE;
}  // AdapterReset

/* -------------------------------------------------------------------------- */

/*
    MiniportPause

Description:
    When a miniport receives a pause request, it enters into a Pausing state.
    The miniport should not indicate up any more network data.  Any pending
    send requests must be completed, and new requests must be rejected with
    NDIS_STATUS_PAUSED.

    Once all sends have been completed and all recieve NBLs have returned to
    the miniport, the miniport enters the Paused state.

    While paused, the miniport can still service interrupts from the hardware
    (to, for example, continue to indicate NDIS_STATUS_MEDIA_CONNECT
    notifications).

    The miniport must continue to be able to handle status indications and OID
    requests.  MiniportPause is different from MiniportHalt because, in
    general, the MiniportPause operation won't release any resources.
    MiniportPause must not attempt to acquire any resources where allocation
    can fail, since MiniportPause itself must not fail.

    MiniportPause runs at IRQL = PASSIVE_LEVEL.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNDIS_MINIPORT_PAUSE_PARAMETERS MiniportPauseParameters
        Additional information about the pause operation.

Return (NDIS_STATUS):
    If the miniport is able to immediately enter the Paused state, it should
    return NDIS_STATUS_SUCCESS.

    If the miniport must wait for send completions or pending receive NBLs, it
    should return NDIS_STATUS_PENDING now, and call NDISMPauseComplete when the
    miniport has entered the Paused state.

    No other return value is permitted.  The pause operation must not fail.
*/

NDIS_STATUS MiniportPause (
    IN  NDIS_HANDLE                     hAdapterContext,
    IN  PNDIS_MINIPORT_PAUSE_PARAMETERS MiniportPauseParameters)
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    PHARDWARE pHardware = &pAdapter->Hardware;
    NDIS_STATUS Status;
#ifdef DBG
    PCHAR reason = NULL;
#endif

    PAGED_CODE();

#ifdef DBG
    switch (MiniportPauseParameters->PauseReason)
    {
        #undef MAKECASE
        #define MAKECASE(pause) case pause: reason = #pause "\n"; break;

        MAKECASE(NDIS_PAUSE_NDIS_INTERNAL)
        MAKECASE(NDIS_PAUSE_LOW_POWER)
        MAKECASE(NDIS_PAUSE_BIND_PROTOCOL)
        MAKECASE(NDIS_PAUSE_UNBIND_PROTOCOL)
        MAKECASE(NDIS_PAUSE_ATTACH_FILTER)
        MAKECASE(NDIS_PAUSE_DETACH_FILTER)
        MAKECASE(NDIS_PAUSE_FILTER_RESTART_STACK)
        MAKECASE(NDIS_PAUSE_MINIPORT_DEVICE_REMOVE)
    }
    if (reason)
    {
        DEBUGP(MP_LOUD, ("%s", reason));
    }
#endif
    pAdapter->PauseReason = MiniportPauseParameters->PauseReason;
    switch (pAdapter->PauseReason)
    {
        case NDIS_PAUSE_ATTACH_FILTER:
            break;
        case NDIS_PAUSE_UNBIND_PROTOCOL:
            if (AcquireAdapter(pAdapter, TRUE))
            {
                HardwareStopReceive(pHardware);
                ReleaseAdapter(pAdapter);
            }
            break;
        case NDIS_PAUSE_DETACH_FILTER:
            break;
        case NDIS_PAUSE_MINIPORT_DEVICE_REMOVE:
#ifdef DBG
            pAdapter->bCheckCounter = TRUE;
            DBG_PRINT("tx ucast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_DIRECTED_FRAMES_XMIT));
            DBG_PRINT("tx mcast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_MULTICAST_FRAMES_XMIT));
            DBG_PRINT("tx bcast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_BROADCAST_FRAMES_XMIT));
            DBG_PRINT("tx bytes: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_BYTES_XMIT));
            DBG_PRINT("tx ucast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_DIRECTED_BYTES_XMIT));
            DBG_PRINT("tx mcast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_MULTICAST_BYTES_XMIT));
            DBG_PRINT("tx bcast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_BROADCAST_BYTES_XMIT));
            DBG_PRINT("rx ucast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_DIRECTED_FRAMES_RCV));
            DBG_PRINT("rx bcast: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_BROADCAST_FRAMES_RCV));
            DBG_PRINT("rx bytes: %llu\n",
                AdapterReadCounter(pAdapter, OID_GEN_BYTES_RCV));
#endif
            break;
        case NDIS_PAUSE_LOW_POWER:
            break;
        default:
#ifdef DBG
            DBG_PRINT("pause: %x\n", pAdapter->PauseReason);
#endif
            break;
    }
#if 0
    {
        //
        // The adapter is busy sending or receiving data, so the pause must
        // be completed asynchronously later.
        //
        pAdapter->ulDriverState |= DRIVER_STATE_PAUSE;
        SetTimer(pAdapter->hResetTimer, 200, FALSE);
        Status = NDIS_STATUS_PENDING;
    }
#endif

    //
    // The pause operation has completed synchronously.
    //
    pAdapter->ulDriverState |= DRIVER_STATE_PAUSED;
    Status = NDIS_STATUS_SUCCESS;

    return Status;
}  // MiniportPause


/*
    MiniportRestart

Description:
    When a miniport receives a restart request, it enters into a Restarting
    state.  The miniport may begin indicating received data (e.g., using
    NdisMIndicateReceiveNetBufferLists), handling status indications, and
    processing OID requests in the Restarting state.  However, no sends will be
    requested while the miniport is in the Restarting state.

    Once the miniport is ready to send data, it has entered the Running state.
    The miniport informs NDIS that it is in the Running state by returning
    NDIS_STATUS_SUCCESS from this MiniportRestart function; or if this function
    has already returned NDIS_STATUS_PENDING, by calling NdisMRestartComplete.

    MiniportRestart runs at IRQL = PASSIVE_LEVEL.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNDIS_MINIPORT_RESTART_PARAMETERS RestartParameters
        Additional information about the restart operation.

Return (NDIS_STATUS):
    If the miniport is able to immediately enter the Running state, it should
    return NDIS_STATUS_SUCCESS.

    If the miniport is still in the Restarting state, it should return
    NDIS_STATUS_PENDING now, and call NdisMRestartComplete when the miniport
    has entered the Running state.

    Other NDIS_STATUS codes indicate errors.  If an error is encountered, the
    miniport must return to the Paused state (i.e., stop indicating receives).
*/

NDIS_STATUS MiniportRestart (
    IN  NDIS_HANDLE                       hAdapterContext,
    IN  PNDIS_MINIPORT_RESTART_PARAMETERS RestartParameters )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    NDIS_STATUS Status = NDIS_STATUS_PENDING;
    PNDIS_RESTART_ATTRIBUTES Attributes = RestartParameters->RestartAttributes;

#if 0
    UNREFERENCED_PARAMETER(RestartParameters);
#endif
    PAGED_CODE();

    while (Attributes)
    {
        if (OID_GEN_MINIPORT_RESTART_ATTRIBUTES == Attributes->Oid)
        {
            PNDIS_RESTART_GENERAL_ATTRIBUTES General =
                (PNDIS_RESTART_GENERAL_ATTRIBUTES) Attributes->Data;

#ifdef DBG
            if (Attributes->DataLength >=
                    sizeof(NDIS_RESTART_GENERAL_ATTRIBUTES))
            {
                DBG_PRINT("mtu: %u\n", General->MtuSize);
                DBG_PRINT("speed: %llu\n", General->MaxXmitLinkSpeed);
                DBG_PRINT("lookahead: %u\n", General->LookaheadSize);
            }
#endif
        }
        Attributes = Attributes->Next;
    }
    pAdapter->ulDriverState &= ~DRIVER_STATE_PAUSED;

#if 0
    //
    // The simulated hardware is immediately ready to send data again, so in
    // this sample code, MiniportRestart always returns success.  If we had to
    // wait for hardware to reinitialize, we'd return NDIS_STATUS_PENDING now,
    // and call NdisMRestartComplete later.
    //
    {
        pAdapter->ulDriverState |= DRIVER_STATE_RESTART;
        SetTimer(pAdapter->hResetTimer, 200, FALSE);
        Status = NDIS_STATUS_PENDING;
    }
#endif
    Status = NDIS_STATUS_SUCCESS;

    return Status;
}  // MiniportRestart


/*
    MiniportResetEx

Description:
    This function is used by NDIS to reset the adapter.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PBOOLEAN AddresssingReset
        Flag to indicate adapter needs the addressing information reloaded.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise, an error code indicating
    failure.
*/

NDIS_STATUS MiniportResetEx (
    IN  NDIS_HANDLE hAdapterContext,
    OUT PBOOLEAN    AddressingReset )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;

    *AddressingReset = FALSE;
    if (AdapterReset(pAdapter))
    {
        pAdapter->ulDriverState &= ~DRIVER_STATE_HANGED;
        pAdapter->ulDriverState |= DRIVER_STATE_RESET;
        SetTimer(pAdapter->hResetTimer, 2000, FALSE);
        return NDIS_STATUS_PENDING;
    }
    return NDIS_STATUS_FAILURE;
}  // MiniportResetEx

/* -------------------------------------------------------------------------- */

/*
    MiniportCheckForHangEx

Description:
    This function is used by NDIS to check if the adapter is not working.
    The NDIS library calls this function every 2 seconds.  If TRUE is
    returned, it will call MiniportResetEx to force an adapter reset.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

Return (BOOLEAN):
    TRUE if adapter requires a reset; FALSE if adapter is functioning
    properly.
*/

BOOLEAN MiniportCheckForHangEx (
    IN  NDIS_HANDLE hAdapterContext )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;

    return AdapterCheckForHang(pAdapter);
}  // MiniportCheckForHangEx

/* -------------------------------------------------------------------------- */

/*
    MiniportDisableInterruptEx

Description:
    This routine is used by NDIS to disable hardware interrupts.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

Return (None):
*/

VOID MiniportDisableInterruptEx (
    IN  NDIS_HANDLE hInterruptContext )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hInterruptContext;

#ifdef DEBUG_INTERRUPT
    DbgPrint("dis int"NEWLINE);
#endif
    HardwareDisableInterrupt(&pAdapter->Hardware);
}  // MiniportDisableInterruptEx


/*
    MiniportEnableInterruptEx

    Description:
        This routine is used by NDIS to enable hardware interrupts.

    Parameters:
        NDIS_HANDLE hAdapaterContext
            Handle to adapter context containing adapter information.

    Return (None):
*/

VOID MiniportEnableInterruptEx (
    IN  NDIS_HANDLE hInterruptContext )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hInterruptContext;

#ifdef DEBUG_INTERRUPT
    DbgPrint("ena int"NEWLINE);
#endif
    HardwareEnableInterrupt(&pAdapter->Hardware);
}  // MiniportEnableInterruptEx

/* -------------------------------------------------------------------------- */

#pragma pack(1)
typedef struct
{
    // Total command length
    UINT   dwLength;

    // Driver Extension Command
    USHORT wCommand;

    // Option flags
    USHORT wFlags;

    // Driver Extension Return Code
    UINT   dwReturnCode;

    // Parameters
    BYTE   bParam[6];
} TDeviceInfo, *PTDeviceInfo;
#pragma pack()


#define DEVICE_COMMAND_INIT       0
#define DEVICE_COMMAND_GET        1
#define DEVICE_COMMAND_SET        2
#define DEVICE_FLAG_OFF           0
#define DEVICE_FLAG_CABLE_STATUS  1
#define DEVICE_FLAG_LINK_STATUS   2
#define DEVICE_FLAG_EEPROM_REG    3

#define DEVICE_FLAG_CAPABILITIES  1

#define DEVICE_COMMAND_COUNTERS   3

#define DEVICE_INIT_SIZE          (sizeof(TDeviceInfo) + 8)
#define DEVICE_INIT_RETURN_SIZE   (sizeof(TDeviceInfo) + 4)
#define DEVICE_GET_SIZE           (sizeof(TDeviceInfo) + 0)
#define DEVICE_CABLE_STATUS_SIZE  (sizeof(TDeviceInfo) + 10 * sizeof(UINT))
#define DEVICE_LINK_STATUS_SIZE   (sizeof(TDeviceInfo) + 6 * sizeof(UINT))
#define DEVICE_COUNTERS_SIZE      (sizeof(TDeviceInfo) + 4 * sizeof(UINT))


NDIS_STATUS GetDeviceMode (
    IN  PNDIS_ADAPTER pAdapter,
    IN  NDIS_OID      Oid,
    IN  PVOID         pBuffer,
    IN  ULONG         ulLen,
    OUT PULONG        pulBytesWritten,
    OUT PULONG        pulBytesNeeded )
{
    NDIS_STATUS status = NDIS_STATUS_SUCCESS;
    PHARDWARE   pHardware = &pAdapter->Hardware;

    // make sure input buffer is valid
    if (pBuffer && ulLen >= sizeof(TDeviceInfo))
    {
        PTDeviceInfo pInfo = (PTDeviceInfo) pBuffer;

        *pulBytesNeeded = 0;
        *pulBytesWritten = sizeof(TDeviceInfo);
        if (Oid == pInfo->dwReturnCode)
        {
            if (DEVICE_COMMAND_GET == pInfo->wCommand)
            {
                switch (pInfo->wFlags)
                {
                    case DEVICE_FLAG_CABLE_STATUS:
                        if (ulLen >= DEVICE_CABLE_STATUS_SIZE)
                        {
                            AcquireAdapter(pAdapter, TRUE);
                            HardwareGetCableStatus(pHardware, pInfo->bParam[5],
                                pInfo->bParam);
                            ReleaseAdapter(pAdapter);
                            *pulBytesWritten = DEVICE_CABLE_STATUS_SIZE;
                        }
                        else
                        {
                            *pulBytesNeeded = DEVICE_CABLE_STATUS_SIZE;
                            status = NDIS_STATUS_BUFFER_TOO_SHORT;
                        }
                        break;
                    case DEVICE_FLAG_LINK_STATUS:
                        if (ulLen >= DEVICE_LINK_STATUS_SIZE)
                        {
                            AcquireAdapter(pAdapter, TRUE);
                            HardwareGetLinkStatus(pHardware, pInfo->bParam[5],
                                pInfo->bParam);
                            ReleaseAdapter(pAdapter);
                            *pulBytesWritten = DEVICE_LINK_STATUS_SIZE;
                        }
                        else
                        {
                            *pulBytesNeeded = DEVICE_LINK_STATUS_SIZE;
                            status = NDIS_STATUS_BUFFER_TOO_SHORT;
                        }
                        break;
                    case DEVICE_FLAG_EEPROM_REG:
                    {
                        PUSHORT pwParam = (PUSHORT) pInfo->bParam;

                        AcquireAdapter(pAdapter, TRUE);
                        pwParam[1] = EepromReadWord(pHardware,
                            (UCHAR) pwParam[0]);
                        ReleaseAdapter(pAdapter);
                        *pulBytesWritten = sizeof(TDeviceInfo);
                        break;
                    }
                    default:
                        status = NDIS_STATUS_INVALID_DATA;
                }
            }
            else if (DEVICE_COMMAND_INIT == pInfo->wCommand)
            {
                if (ulLen >= DEVICE_INIT_SIZE)
                {
                    pInfo->bParam[0] = 'M';
                    pInfo->bParam[1] = 'i';
                    pInfo->bParam[2] = 'c';
                    pInfo->bParam[3] = 'r';
                    //DAVIDCAI added 12-17-2009, using bparam[4] to pass back the port information
                    //to new Mirecl LinkMD driver.
                    #ifdef DEF_KS8842
                      pInfo->bParam[4]=2;
                    #else
                     pInfo->bParam[4]=1;
                    #endif
                    //DAVIDCAI end
                    *pulBytesWritten = DEVICE_INIT_RETURN_SIZE;
                }
                else
                {
                    *pulBytesNeeded = DEVICE_INIT_SIZE;
                    status = NDIS_STATUS_BUFFER_TOO_SHORT;
                }
            }
            else
            {
                status = NDIS_STATUS_INVALID_DATA;
            }
        }
        else
        {
            status = NDIS_STATUS_INVALID_DATA;
        }
        pInfo->dwLength = *pulBytesWritten;
        pInfo->dwReturnCode = status;
    }
    else
    {
        *pulBytesNeeded = sizeof(TDeviceInfo);
        *pulBytesWritten = 0;
        status = NDIS_STATUS_BUFFER_TOO_SHORT;
    }
    return status;
}  // GetDeviceMode


NDIS_STATUS SetDeviceMode (
    IN  PNDIS_ADAPTER pAdapter,
    IN  NDIS_OID      Oid,
    IN  PVOID         pBuffer,
    IN  ULONG         ulLen,
    OUT PULONG        pulBytesRead,
    OUT PULONG        pulBytesNeeded )
{
    NDIS_STATUS status = NDIS_STATUS_SUCCESS;
    PHARDWARE   pHardware = &pAdapter->Hardware;

    if (pBuffer && ulLen >= sizeof(TDeviceInfo))
    {
        PTDeviceInfo pInfo = (PTDeviceInfo) pBuffer;

        *pulBytesNeeded = 0;
        *pulBytesRead = sizeof(TDeviceInfo);
        if (Oid == pInfo->dwReturnCode &&
                DEVICE_COMMAND_SET == pInfo->wCommand)
        {
            switch (pInfo->wFlags)
            {
                case DEVICE_FLAG_CAPABILITIES:
                {
                    PUINT pulData = (PUINT) pInfo->bParam;

                    AcquireAdapter(pAdapter, TRUE);
                    HardwareSetCapabilities(pHardware, 0, *pulData);
                    ReleaseAdapter(pAdapter);
                    break;
                }
                case DEVICE_FLAG_EEPROM_REG:
                {
                    PUSHORT pwParam = (PUSHORT) pInfo->bParam;

                    AcquireAdapter(pAdapter, TRUE);
                    EepromWriteWord(pHardware, (UCHAR) pwParam[0],
                        pwParam[1]);
                    ReleaseAdapter(pAdapter);
                    break;
                }
                default:
                    status = NDIS_STATUS_INVALID_DATA;
            }
        }
        else
            status = NDIS_STATUS_INVALID_DATA;
        pInfo->dwLength = *pulBytesRead;
        pInfo->dwReturnCode = status;
    }
    else
    {
        *pulBytesNeeded = sizeof(TDeviceInfo);
        *pulBytesRead = 0;
        status = NDIS_STATUS_BUFFER_TOO_SHORT;
    }
    return status;
}  // SetDeviceMode


BOOLEAN PciReadConfig (
    IN  PVOID  pAdapterContext,
    IN  int    offset,
    OUT PUINT  pulData )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) pAdapterContext;

    if (NdisMGetBusData(pAdapter->hAdapter, PCI_WHICHSPACE_CONFIG,
            offset, pulData, sizeof(UINT)) == sizeof(UINT))
        return TRUE;
    return FALSE;
}  /* PciReadConfig */


BOOLEAN PciReadConfigByte (
    IN  PVOID  pAdapterContext,
    IN  int    offset,
    OUT PUCHAR pbData )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) pAdapterContext;

    if (NdisMGetBusData(pAdapter->hAdapter, PCI_WHICHSPACE_CONFIG,
            offset, pbData, sizeof(UCHAR)) == sizeof(UCHAR))
        return TRUE;
    return FALSE;
}  /* PciReadConfigByte */


BOOLEAN PciReadConfigWord (
    IN  PVOID   pAdapterContext,
    IN  int     offset,
    OUT PUSHORT pwData )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) pAdapterContext;

    if (NdisMGetBusData(pAdapter->hAdapter, PCI_WHICHSPACE_CONFIG,
            offset, pwData, sizeof(USHORT)) == sizeof(USHORT))
        return TRUE;
    return FALSE;
}  /* PciReadConfigWord */


BOOLEAN PciWriteConfig (
    IN  PVOID pAdapterContext,
    IN  int   offset,
    IN  UINT  ulData )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) pAdapterContext;

    if (NdisMSetBusData(pAdapter->hAdapter, PCI_WHICHSPACE_CONFIG,
            offset, &ulData, sizeof(UINT)) == sizeof(UINT))
        return TRUE;
    return FALSE;
}  /* PciWriteConfig */


BOOLEAN PciWriteConfigByte (
    IN  PVOID pAdapterContext,
    IN  int   offset,
    IN  UCHAR bData )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) pAdapterContext;

    if (NdisMSetBusData(pAdapter->hAdapter, PCI_WHICHSPACE_CONFIG,
            offset, &bData, sizeof(UCHAR)) == sizeof(UCHAR))
        return TRUE;
    return FALSE;
}  /* PciWriteConfigByte */


BOOLEAN PciWriteConfigWord (
    IN  PVOID  pAdapterContext,
    IN  int    offset,
    IN  USHORT wData )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) pAdapterContext;

    if (NdisMSetBusData(pAdapter->hAdapter, PCI_WHICHSPACE_CONFIG,
            offset, &wData, sizeof(USHORT)) == sizeof(USHORT))
        return TRUE;
    return FALSE;
}  /* PciWriteConfigWord */
